#ifndef RATE_ALLOCATOR_H__
#define RATE_ALLOCATOR_H__

#include <stddef.h>
#include <stdint.h>
#include "linelet.h"
#include "rate_allocator.h"

#ifdef __cplusplus
extern "C" {
#endif

struct linelet_rate_allocator_quant
{
   int8_t quant[LINELET_MAX_LINES][LINELET_MAX_COMPONENTS][LINELET_MAX_BANDS];
};

// For SIMD optimization purposes, this must be exactly 64-bits large and laid out like this.
struct linelet_precinct_range_level
{
   int16_t lo;
   int16_t hi;
   uint16_t bytes_total;
   uint16_t bits;
};

typedef struct linelet_rate_allocator linelet_rate_allocator_t;

linelet_rate_allocator_t *linelet_rate_allocator_new(unsigned pixels_per_precinct, unsigned samples_per_line, unsigned lines,
      const struct linelet_component *components, unsigned num_components,
      const struct linelet_rate_allocator_quant *quant, unsigned num_quant);

#define LINELET_RATE_ALLOCATOR_NO_QUANT -1
void linelet_rate_allocator_minimum_quant(linelet_rate_allocator_t *ra, int q);

void linelet_rate_allocator_set_quant(linelet_rate_allocator_t *ra,
      const struct linelet_rate_allocator_quant *quant, unsigned num_quant);

void linelet_rate_allocator_study_lines(linelet_rate_allocator_t *ra,
      const struct linelet_wavelet_buffer *buffers, unsigned target_buffer_size);

unsigned linelet_rate_allocator_quant_factor(linelet_rate_allocator_t *ra, unsigned line, unsigned component, unsigned band);
unsigned linelet_rate_allocator_bits(linelet_rate_allocator_t *ra, unsigned line, unsigned component, unsigned precinct, unsigned band);
const struct linelet_precinct_range_level *linelet_rate_allocator_precincts(linelet_rate_allocator_t *ra,
      unsigned line, unsigned component, unsigned band);

void linelet_rate_allocator_free(linelet_rate_allocator_t *ra);

#ifdef __cplusplus
}
#endif
#endif

