#ifndef RPNG_H__
#define RPNG_H__

#include <stdint.h>
#include <stddef.h>

#ifndef __cplusplus
#include <stdbool.h>
#endif

#ifdef HAVE_CONFIG_H
#include "../../config.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

// Control how RPNG allocates memory.
void rpng_set_malloc(void *(*cb)(size_t));
void rpng_set_calloc(void *(*cb)(size_t, size_t));
void rpng_set_free(void (*cb)(void*));

// Image loading routines.
// A8R8G8B8 packed in a uint32_t.
bool rpng_load_image_argb(const char *path, uint32_t **data, unsigned *width, unsigned *height);
// A16R16G16B16 packed in a uint64_t.
bool rpng_load_image_argb16(const char *path, uint64_t **data, unsigned *width, unsigned *height);
// R16G16B16 packed in consecutive uint16_t elements.
bool rpng_load_image_a2bgr10(const char *path, uint32_t **data, unsigned *width, unsigned *height);

// Image saving routines. Compression is zlib compression, and has a range of 0 (no compression) to 9 (slowest, best compression).
#define RPNG_COMPRESSION_MINIMUM 0
#define RPNG_COMPRESSION_MAXIMUM 9
bool rpng_save_image_argb16(const char *path, const uint64_t *data,
      unsigned width, unsigned, unsigned pitch, unsigned compression);
bool rpng_save_image_a2bgr10(const char *path, const uint32_t *data,
      unsigned width, unsigned, unsigned pitch, unsigned compression);
bool rpng_save_image_argb(const char *path, const uint32_t *data,
      unsigned width, unsigned height, unsigned pitch, unsigned compression);
// B8G8R8 packed in consecutive bytes.
bool rpng_save_image_bgr24(const char *path, const uint8_t *data,
      unsigned width, unsigned height, unsigned pitch, unsigned compression);

#ifdef __cplusplus
}
#endif

#endif

