#ifndef LINELET_UTIL_H__
#define LINELET_UTIL_H__

#include <limits.h>
#include <stdint.h>
#ifdef __cplusplus
extern "C" {
#endif

// __builtin_clz of 0 is undefined.
#ifdef __GNUC__
static inline unsigned linelet_clz(unsigned v)
{
   return __builtin_clz(v);
}
#else
static inline unsigned linelet_clz(unsigned v)
{
   unsigned leading = 0;
   v = ~v;
   while (v & 0x80000000u)
   {
      leading++;
      v <<= 1;
   }
   return leading;
}
#endif

static inline unsigned linelet_log2i(unsigned v)
{
   return (sizeof(unsigned) * CHAR_BIT - 1) - linelet_clz(v);
}

static inline void linelet_write64_be(uint8_t *buf, uint64_t v)
{
   for (unsigned i = 0; i < 8; i++)
   {
      buf[i] = (v >> 56) & 0xff;
      v <<= 8;
   }
}

static inline void linelet_write32_be(uint8_t *buf, uint32_t v)
{
   buf[0] = (v >> 24) & 0xff;
   buf[1] = (v >> 16) & 0xff;
   buf[2] = (v >>  8) & 0xff;
   buf[3] = (v >>  0) & 0xff;
}

static inline void linelet_write16_be(uint8_t *buf, uint16_t v)
{
   buf[0] = (v >> 8) & 0xff;
   buf[1] = (v >> 0) & 0xff;
}

static inline uint64_t linelet_read64_be(const uint8_t *buf)
{
   uint64_t ret = 0;
   for (unsigned i = 0; i < 8; i++)
   {
      ret <<= 8;
      ret |= buf[i];
   }
   return ret;
}

static inline uint32_t linelet_read32_be(const uint8_t *buf)
{
   return (buf[0] << 24u) | (buf[1] << 16u) | (buf[2] << 8u) | buf[3];
}

static inline uint16_t linelet_read16_be(const uint8_t *buf)
{
   return (buf[0] << 8u) | buf[1];
}

#define LLMAX(a, b) ((a) > (b) ? (a) : (b))
#define LLMIN(a, b) ((a) < (b) ? (a) : (b))

#ifdef __cplusplus
}
#endif
#endif

