#include <stdint.h>

void linelet_prefilter_line2_c(int16_t *l0, int16_t *l1, unsigned samples)
{
   for (unsigned x = 0; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];

      // Pre
      s1 = s0 - s1;
      s0 -= s1 >> 1;

      // V matrix
      s1 <<= 1;

      // Post
      s0 += s1 >> 1;
      s1 = s0 - s1;

      l0[x] = s0;
      l1[x] = s1;
   }
}

void linelet_postfilter_line2_c(int16_t *l0, int16_t *l1, unsigned samples)
{
   for (unsigned x = 0; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];

      // Inverse post
      s1 = s0 - s1;
      s0 -= s1 >> 1;

      // V^-1
      s1 >>= 1;

      // Inverse pre
      s0 += s1 >> 1;
      s1 = s0 - s1;

      l0[x] = s0;
      l1[x] = s1;
   }
}

void linelet_prefilter_line4_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples)
{
   for (unsigned x = 0; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];

      s2 = s1 - s2;
      s1 -= s2 >> 1;

      s2 <<= 1;
      s2 += (s0 - s3) >> 1;

      s1 += s2 >> 1;
      s2 = s1 - s2;

      l1[x] = s1;
      l2[x] = s2;
   }
}

void linelet_postfilter_line4_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples)
{
   for (unsigned x = 0; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];

      s2 = s1 - s2;
      s1 -= s2 >> 1;

      s2 -= (s0 - s3) >> 1;
      s2 >>= 1;

      s1 += s2 >> 1;
      s2 = s1 - s2;

      l1[x] = s1;
      l2[x] = s2;
   }
}

void linelet_prefilter_line8_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3,
      int16_t *l4, int16_t *l5, int16_t *l6, int16_t *l7,
      unsigned samples)
{
   for (unsigned x = 0; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];
      int16_t s4 = l4[x];
      int16_t s5 = l5[x];
      int16_t s6 = l6[x];
      int16_t s7 = l7[x];

      s6 = s1 - s6;
      s1 -= s6 >> 1;
      s5 = s2 - s5;
      s2 -= s5 >> 1;
      s4 = s3 - s4;
      s3 -= s4 >> 1;

      s4 <<= 1;
      s6 += (s0 - s7) >> 3;
      s5 += s6 >> 2;
      s4 += s5 >> 1;

      s3 += s4 >> 1;
      s4 = s3 - s4;
      s2 += s5 >> 1;
      s5 = s2 - s5;
      s1 += s6 >> 1;
      s6 = s1 - s6;

      l1[x] = s1;
      l2[x] = s2;
      l3[x] = s3;
      l4[x] = s4;
      l5[x] = s5;
      l6[x] = s6;
   }
}

void linelet_postfilter_line8_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3,
      int16_t *l4, int16_t *l5, int16_t *l6, int16_t *l7,
      unsigned samples)
{
   for (unsigned x = 0; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];
      int16_t s4 = l4[x];
      int16_t s5 = l5[x];
      int16_t s6 = l6[x];
      int16_t s7 = l7[x];

      s6 = s1 - s6;
      s1 -= s6 >> 1;
      s5 = s2 - s5;
      s2 -= s5 >> 1;
      s4 = s3 - s4;
      s3 -= s4 >> 1;

      s4 -= s5 >> 1;
      s5 -= s6 >> 2;
      s6 -= (s0 - s7) >> 3;
      s4 >>= 1;

      s3 += s4 >> 1;
      s4 = s3 - s4;
      s2 += s5 >> 1;
      s5 = s2 - s5;
      s1 += s6 >> 1;
      s6 = s1 - s6;

      l1[x] = s1;
      l2[x] = s2;
      l3[x] = s3;
      l4[x] = s4;
      l5[x] = s5;
      l6[x] = s6;
   }
}

