#include "simd.h"
#include <string.h>
struct linelet_simd_funcs linelet_simd;

// C routines
// RCT
void linelet_rct_forward_argb8_yuv8_line_c(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_inverse_yuv8_argb8_line_c(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

void linelet_rct_forward_a2bgr10_yuv10_line_c(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_inverse_yuv10_a2bgr10_line_c(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

void linelet_rct_forward_gbrp8_line_c(int16_t *l, int16_t *cb, int16_t *cr, const uint8_t *g, const uint8_t *b, const uint8_t *r, unsigned samples);
void linelet_rct_inverse_gbrp8_line_c(uint8_t *g, uint8_t *b, uint8_t *r, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);
void linelet_rct_forward_gbrp10_line_c(int16_t *l, int16_t *cb, int16_t *cr, const uint16_t *g, const uint16_t *b, const uint16_t *r, unsigned samples);
void linelet_rct_inverse_gbrp10_line_c(uint16_t *g, uint16_t *b, uint16_t *r, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

// BT
void linelet_bt_yuv8_forward_line_c(int16_t *dst, const uint8_t *src, unsigned samples);
void linelet_bt_yuv8_inverse_line_c(uint8_t *dst, const int16_t *src, unsigned samples);

void linelet_bt_yuv10_forward_line_c(int16_t *dst, const uint16_t *src, unsigned samples);
void linelet_bt_yuv10_inverse_line_c(uint16_t *dst, const int16_t *src, unsigned samples);

void linelet_bt_yuv12_forward_line_c(int16_t *dst, const uint16_t *src, unsigned samples);
void linelet_bt_yuv12_inverse_line_c(uint16_t *dst, const int16_t *src, unsigned samples);

// Haar
void linelet_haar_line_forward_c(int16_t *a, int16_t *b, unsigned samples);
void linelet_haar_line_inverse_c(int16_t *a, int16_t *b, unsigned samples);

// Le Gall
void linelet_legall_line_forward_c(int16_t *lp, int16_t *hp, const int16_t *in, unsigned samples);
void linelet_legall_line_inverse_c(int16_t *out, const int16_t *lp, const int16_t *hp, unsigned samples);

void linelet_legall_vert_line_forward_top_c(
      int16_t *lp0,
      int16_t *hp0,
      int16_t *lp1,
      unsigned samples);

void linelet_legall_vert_line_forward_c(
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      int16_t *lp2,
      unsigned samples);

void linelet_legall_vert_line_forward_bottom_c(
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      unsigned samples);

void linelet_legall_vert_line_inverse_top_c(
      int16_t *lp0,
      int16_t *hp0,
      unsigned samples);

void linelet_legall_vert_line_inverse_c(
      int16_t *lp0,
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      unsigned samples);

void linelet_legall_vert_line_inverse_bottom_c(
      int16_t *lp0,
      int16_t *hp0,
      unsigned samples);

// Prepost filter
void linelet_prefilter_line2_c(int16_t *l0, int16_t *l1, unsigned samples);
void linelet_postfilter_line2_c(int16_t *l0, int16_t *l1, unsigned samples);
void linelet_prefilter_line4_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples);
void linelet_postfilter_line4_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples);
void linelet_prefilter_line8_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, int16_t *l4, int16_t *l5, int16_t *l6, int16_t *l7, unsigned samples);
void linelet_postfilter_line8_c(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, int16_t *l4, int16_t *l5, int16_t *l6, int16_t *l7, unsigned samples);

// Range
void linelet_range_min_max_line_8_c(struct linelet_precinct_range_level *ranges, const int16_t *src, unsigned precincts, unsigned step);
void linelet_range_min_max_line_4_c(struct linelet_precinct_range_level *ranges, const int16_t *src, unsigned precincts);

// Quantizer
void linelet_quantize_line_c(int16_t *data, unsigned samples, unsigned q, int16_t adder);
void linelet_dequantize_line_c(int16_t *data, unsigned samples, unsigned q);

// Packing
#define PACK16(v, inst) uint8_t *linelet_pack_line16_##v##_##inst(uint8_t *output, const int16_t *input, unsigned samples)
#define PACK8(v, inst) uint8_t *linelet_pack_line8_##v##_##inst(uint8_t *output, const int16_t *input)
#define PACK4(v, inst) uint8_t *linelet_pack_line4_##v##_##inst(uint8_t *output, const int16_t *input)
#define UNPACK8(v, inst) const uint8_t *linelet_unpack_line8_##v##_##inst(int16_t *output, const uint8_t *input, unsigned samples, const uint64_t *lut)
#define UNPACK4(v, inst) const uint8_t *linelet_unpack_line4_##v##_##inst(int16_t *output, const uint8_t *input, const uint64_t *lut)

#define STAMP_DECL(func, inst) \
   func(1, inst); func(2, inst); func(3, inst); func(4, inst); \
   func(5, inst); func(6, inst); func(7, inst); func(8, inst); \
   func(9, inst); func(10, inst); func(11, inst); func(12, inst); \
   func(13, inst); func(14, inst); func(15, inst); func(16, inst)

STAMP_DECL(PACK16,  c);
STAMP_DECL(PACK8,   c);
STAMP_DECL(PACK4,   c);
STAMP_DECL(UNPACK8, c);
STAMP_DECL(UNPACK4, c);

#define FUNC(fun, len, v, inst) linelet_##fun##_line##len##_##v##_##inst

#define STAMP_TABLE(func, len, inst) \
   FUNC(func, len,  1, inst), \
   FUNC(func, len,  2, inst), \
   FUNC(func, len,  3, inst), \
   FUNC(func, len,  4, inst), \
   FUNC(func, len,  5, inst), \
   FUNC(func, len,  6, inst), \
   FUNC(func, len,  7, inst), \
   FUNC(func, len,  8, inst), \
   FUNC(func, len,  9, inst), \
   FUNC(func, len, 10, inst), \
   FUNC(func, len, 11, inst), \
   FUNC(func, len, 12, inst), \
   FUNC(func, len, 13, inst), \
   FUNC(func, len, 14, inst), \
   FUNC(func, len, 15, inst), \
   FUNC(func, len, 16, inst),

static const linelet_pack16_t pack16_c[16] = {
   STAMP_TABLE(pack, 16, c)
};

static const linelet_pack8_t pack8_c[16] = {
   STAMP_TABLE(pack, 8, c)
};

static const linelet_pack4_t pack4_c[16] = {
   STAMP_TABLE(pack, 4, c)
};

static const linelet_unpack8_t unpack8_c[16] = {
   STAMP_TABLE(unpack, 8, c)
};

static const linelet_unpack4_t unpack4_c[16] = {
   STAMP_TABLE(unpack, 4, c)
};

static void emms_dummy(void)
{}

#undef INIT
#define INIT(func) linelet_simd.func = linelet_##func##_c
void linelet_simd_init_c(void)
{
   linelet_simd.emms = emms_dummy;

   INIT(prefilter_line2);
   INIT(postfilter_line2);
   INIT(prefilter_line4);
   INIT(postfilter_line4);
   INIT(prefilter_line8);
   INIT(postfilter_line8);

   INIT(rct_forward_argb8_yuv8_line);
   INIT(rct_forward_a2bgr10_yuv10_line);
   INIT(rct_inverse_yuv8_argb8_line);
   INIT(rct_inverse_yuv10_a2bgr10_line);

   INIT(rct_forward_gbrp8_line);
   INIT(rct_inverse_gbrp8_line);
   INIT(rct_forward_gbrp10_line);
   INIT(rct_inverse_gbrp10_line);

   INIT(bt_yuv8_forward_line);
   INIT(bt_yuv8_inverse_line);
   INIT(bt_yuv10_forward_line);
   INIT(bt_yuv10_inverse_line);
   INIT(bt_yuv12_forward_line);
   INIT(bt_yuv12_inverse_line);

   INIT(haar_line_forward);
   INIT(haar_line_inverse);

   INIT(legall_line_forward);
   INIT(legall_line_inverse);
   INIT(legall_vert_line_forward_top);
   INIT(legall_vert_line_forward);
   INIT(legall_vert_line_forward_bottom);
   INIT(legall_vert_line_inverse_top);
   INIT(legall_vert_line_inverse);
   INIT(legall_vert_line_inverse_bottom);

   INIT(range_min_max_line_8);
   INIT(range_min_max_line_4);

   INIT(quantize_line);
   INIT(dequantize_line);

   memcpy(linelet_simd.pack16,  pack16_c,  sizeof(pack16_c));
   memcpy(linelet_simd.pack8,   pack8_c,   sizeof(pack8_c));
   memcpy(linelet_simd.pack4,   pack4_c,   sizeof(pack4_c));
   memcpy(linelet_simd.unpack8, unpack8_c, sizeof(unpack8_c));
   memcpy(linelet_simd.unpack4, unpack4_c, sizeof(unpack4_c));
}

