#include "haar.h"
#include "timer.h"
#include "simd.h"

void linelet_haar_dwt_forward(int16_t * restrict a, int16_t * restrict b, unsigned width)
{
   TIMER_COUNTER(haar_dwt_forward);
   TIMER_START_COUNT(haar_dwt_forward, width);

   unsigned x = 0;
   if (width >= 16)
   {
      unsigned asm_width = width & ~15;
      linelet_simd.haar_line_forward(a, b, asm_width);
      x = asm_width;
   }

   for (; x < width; x++)
   {
      b[x] -= a[x];
      a[x] += (b[x] + 1) >> 1;
   }

   linelet_simd.emms();
   TIMER_STOP(haar_dwt_forward);
}

void linelet_haar_dwt_inverse(int16_t * restrict a, int16_t * restrict b, unsigned width)
{
   TIMER_COUNTER(haar_dwt_inverse);
   TIMER_START_COUNT(haar_dwt_inverse, width);

   unsigned x = 0;
   if (width >= 16)
   {
      unsigned asm_width = width & ~15;
      linelet_simd.haar_line_inverse(a, b, asm_width);
      x = asm_width;
   }

   for (; x < width; x++)
   {
      a[x] -= (b[x] + 1) >> 1;
      b[x] += a[x];
   }

   linelet_simd.emms();
   TIMER_STOP(haar_dwt_inverse);
}

