#include "log.h"
#include <stdio.h>

static linelet_vlog_callback_t log_cb;
static void *log_userdata;
static enum linelet_log_level log_minimum_level;

void linelet_log_set_log_level(enum linelet_log_level level)
{
   log_minimum_level = level;
}

void linelet_log_set_callback(linelet_vlog_callback_t cb, void *userdata)
{
   log_cb = cb;
   log_userdata = userdata;
}

void linelet_log_printf(enum linelet_log_level level, const char *fmt, ...)
{
   if (level < log_minimum_level)
      return;

   va_list va;
   va_start(va, fmt);
   linelet_log_vprintf(level, fmt, va);
   va_end(va);
}

void linelet_log_vprintf(enum linelet_log_level level, const char *fmt, va_list va)
{
   if (level < log_minimum_level)
      return;

   if (log_cb)
      log_cb(log_userdata, level, fmt, va);
   else
   {
      switch (level)
      {
         case LINELET_LOG_DEBUG: fprintf(stderr, "[DEBUG] "); break;
         case LINELET_LOG_INFO: fprintf(stderr, "[INFO] "); break;
         case LINELET_LOG_WARN: fprintf(stderr, "[WARN] "); break;
         case LINELET_LOG_ERROR: fprintf(stderr, "[ERROR] "); break;
         default: break;
      }
      vfprintf(stderr, fmt, va);
   }
}

