#include "prepostfilter.h"
#include "timer.h"
#include "simd.h"

void linelet_prefilter_8(
      int16_t * restrict l0,
      int16_t * restrict l1,
      int16_t * restrict l2,
      int16_t * restrict l3,
      int16_t * restrict l4,
      int16_t * restrict l5,
      int16_t * restrict l6,
      int16_t * restrict l7,
      unsigned samples)
{
   TIMER_COUNTER(prefilter_8);
   TIMER_START_COUNT(prefilter_8, samples);

   linelet_prefilter_line8_t simd = linelet_simd.prefilter_line8;

   unsigned x = 0;
   if (samples >= 8)
   {
      x = samples & ~7;
      simd(l0, l1, l2, l3, l4, l5, l6, l7, x);
   }

   for (; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];
      int16_t s4 = l4[x];
      int16_t s5 = l5[x];
      int16_t s6 = l6[x];
      int16_t s7 = l7[x];

      s6 = s1 - s6;
      s1 -= s6 >> 1;
      s5 = s2 - s5;
      s2 -= s5 >> 1;
      s4 = s3 - s4;
      s3 -= s4 >> 1;

      s4 <<= 1;
      s6 += (s0 - s7) >> 3;
      s5 += s6 >> 2;
      s4 += s5 >> 1;

      s3 += s4 >> 1;
      s4 = s3 - s4;
      s2 += s5 >> 1;
      s5 = s2 - s5;
      s1 += s6 >> 1;
      s6 = s1 - s6;

      l1[x] = s1;
      l2[x] = s2;
      l3[x] = s3;
      l4[x] = s4;
      l5[x] = s5;
      l6[x] = s6;
   }
   linelet_simd.emms();
   TIMER_STOP(prefilter_8);
}

void linelet_postfilter_8(
      int16_t * restrict l0,
      int16_t * restrict l1,
      int16_t * restrict l2,
      int16_t * restrict l3,
      int16_t * restrict l4,
      int16_t * restrict l5,
      int16_t * restrict l6,
      int16_t * restrict l7,
      unsigned samples)
{
   TIMER_COUNTER(postfilter_8);
   TIMER_START_COUNT(postfilter_8, samples);

   linelet_postfilter_line8_t simd = linelet_simd.postfilter_line8;

   unsigned x = 0;
   if (samples >= 8)
   {
      x = samples & ~7;
      simd(l0, l1, l2, l3, l4, l5, l6, l7, x);
   }

   for (; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];
      int16_t s4 = l4[x];
      int16_t s5 = l5[x];
      int16_t s6 = l6[x];
      int16_t s7 = l7[x];

      s6 = s1 - s6;
      s1 -= s6 >> 1;
      s5 = s2 - s5;
      s2 -= s5 >> 1;
      s4 = s3 - s4;
      s3 -= s4 >> 1;

      s4 -= s5 >> 1;
      s5 -= s6 >> 2;
      s6 -= (s0 - s7) >> 3;
      s4 >>= 1;

      s3 += s4 >> 1;
      s4 = s3 - s4;
      s2 += s5 >> 1;
      s5 = s2 - s5;
      s1 += s6 >> 1;
      s6 = s1 - s6;

      l1[x] = s1;
      l2[x] = s2;
      l3[x] = s3;
      l4[x] = s4;
      l5[x] = s5;
      l6[x] = s6;
   }

   linelet_simd.emms();
   TIMER_STOP(postfilter_8);
}

void linelet_prefilter_4(
      int16_t * restrict l0,
      int16_t * restrict l1,
      int16_t * restrict l2,
      int16_t * restrict l3,
      unsigned samples)
{
   TIMER_COUNTER(prefilter_4);
   TIMER_START_COUNT(prefilter_4, samples);

   linelet_prefilter_line4_t simd = linelet_simd.prefilter_line4;

   unsigned x = 0;
   if (samples >= 8)
   {
      x = samples & ~7;
      simd(l0, l1, l2, l3, x);
   }

   for (; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];

      s2 = s1 - s2;
      s1 -= s2 >> 1;

      s2 <<= 1;
      s2 += (s0 - s3) >> 1;

      s1 += s2 >> 1;
      s2 = s1 - s2;

      l1[x] = s1;
      l2[x] = s2;
   }

   linelet_simd.emms();
   TIMER_STOP(prefilter_4);
}

void linelet_postfilter_4(
      int16_t * restrict l0,
      int16_t * restrict l1,
      int16_t * restrict l2,
      int16_t * restrict l3,
      unsigned samples)
{
   TIMER_COUNTER(postfilter_4);
   TIMER_START_COUNT(postfilter_4, samples);

   linelet_postfilter_line4_t simd = linelet_simd.postfilter_line4;

   unsigned x = 0;
   if (samples >= 8)
   {
      x = samples & ~7;
      simd(l0, l1, l2, l3, x);
   }

   for (; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];
      int16_t s2 = l2[x];
      int16_t s3 = l3[x];

      s2 = s1 - s2;
      s1 -= s2 >> 1;

      s2 -= (s0 - s3) >> 1;
      s2 >>= 1;

      s1 += s2 >> 1;
      s2 = s1 - s2;

      l1[x] = s1;
      l2[x] = s2;
   }

   linelet_simd.emms();
   TIMER_STOP(postfilter_4);
}

void linelet_prefilter_2(
      int16_t * restrict l0,
      int16_t * restrict l1,
      unsigned samples)
{
   TIMER_COUNTER(prefilter_2);
   TIMER_START_COUNT(prefilter_2, samples);

   linelet_prefilter_line2_t simd = linelet_simd.prefilter_line2;

   unsigned x = 0;
   if (samples >= 8)
   {
      x = samples & ~7;
      simd(l0, l1, x);
   }

   for (; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];

      // Pre
      s1 = s0 - s1;
      s0 -= s1 >> 1;

      // V matrix
      s1 <<= 1;

      // Post
      s0 += s1 >> 1;
      s1 = s0 - s1;

      l0[x] = s0;
      l1[x] = s1;
   }

   linelet_simd.emms();
   TIMER_STOP(prefilter_2);
}

void linelet_postfilter_2(
      int16_t * restrict l0,
      int16_t * restrict l1,
      unsigned samples)
{
   TIMER_COUNTER(postfilter_2);
   TIMER_START_COUNT(postfilter_2, samples);

   linelet_postfilter_line2_t simd = linelet_simd.postfilter_line2;

   unsigned x = 0;
   if (samples >= 8)
   {
      x = samples & ~7;
      simd(l0, l1, x);
   }

   for (; x < samples; x++)
   {
      int16_t s0 = l0[x];
      int16_t s1 = l1[x];

      // Inverse post
      s1 = s0 - s1;
      s0 -= s1 >> 1;

      // V^-1
      s1 >>= 1;

      // Inverse pre
      s0 += s1 >> 1;
      s1 = s0 - s1;

      l0[x] = s0;
      l1[x] = s1;
   }

   linelet_simd.emms();
   TIMER_STOP(postfilter_2);
}

