#include "quantizer.h"
#include "timer.h"
#include "simd.h"

void linelet_quantize_samples(int16_t *data, size_t samples, unsigned q)
{
   if (!q)
      return;

   TIMER_COUNTER(quantize);
   TIMER_START_COUNT(quantize, samples);

   const int16_t adder = (1 << (q - 1)) - 1;

   unsigned i = 0;
   if (samples >= 32)
   {
      unsigned asm_samples = samples & ~31;
      linelet_simd.quantize_line(data, asm_samples, q, adder);
      i = asm_samples;
      data += asm_samples;
   }

   for (; i < samples; i++, data++)
      *data = (*data + adder) >> q;

   linelet_simd.emms();
   TIMER_STOP(quantize);
}

void linelet_dequantize_samples(int16_t *data, size_t samples, unsigned q)
{
   if (!q)
      return;

   TIMER_COUNTER(dequantize);
   TIMER_START_COUNT(dequantize, samples);

   unsigned i = 0;
   if (samples >= 32)
   {
      unsigned asm_samples = samples & ~31;
      linelet_simd.dequantize_line(data, asm_samples, q);
      i = asm_samples;
      data += asm_samples;
   }

   for (; i < samples; i++)
      *data++ <<= q;

   linelet_simd.emms();
   TIMER_STOP(dequantize);
}

