#include "rate_allocator.h"
#include "range.h"
#include "memory.h"
#include "timer.h"
#include "util.h"
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#define PSY(a, h, hh, hhh) { (a) + h, (a) - 1 + hh, (a) - 2 + hhh, a - 3, a - 4, a - 5, a - 6, a - 7 }
#define PSY_LINE(a, h, hh, hhh, l, c) { [0] = PSY((a) + (l), h, hh, hhh), [1] = PSY((a) + (c), h, hh, hhh), [2] = PSY((a) + (c), h, hh, hhh) }
#define PSY_LUMA_OFFSET 0
#define PSY_CHROMA_OFFSET 1

#define DEFAULT_QUANT_TABLE_ELEMS 16
static const struct linelet_rate_allocator_quant default_quant_table[5][DEFAULT_QUANT_TABLE_ELEMS] = {
   // Line sizes - 1
   [0] = {
#define Q(i) [i] = { .quant = { \
   [0] = PSY_LINE(i, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET) }}
      Q(0), Q(1), Q(2), Q(3), Q(4), Q(5), Q(6), Q(7), Q(8), Q(9), Q(10), Q(11), Q(12), Q(13), Q(14), Q(15),
#undef Q
   },

   [1] = {
#define Q(i) [i] = { .quant = { \
   [0] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [1] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET) }}
      Q(0), Q(1), Q(2), Q(3), Q(4), Q(5), Q(6), Q(7), Q(8), Q(9), Q(10), Q(11), Q(12), Q(13), Q(14), Q(15),
#undef Q
   },

   [2] = {
#define Q(i) [i] = { .quant = { \
   [0] = PSY_LINE(i - 2, 1, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [1] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [2] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [3] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET) }}
      Q(0), Q(1), Q(2), Q(3), Q(4), Q(5), Q(6), Q(7), Q(8), Q(9), Q(10), Q(11), Q(12), Q(13), Q(14), Q(15),
#undef Q
   },

   [3] = {
#define Q(i) [i] = { .quant = { \
   [0] = PSY_LINE(i - 3, 2, 1, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [1] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [2] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [3] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [4] = PSY_LINE(i - 2, 1, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [5] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [6] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [7] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET) }}
      Q(0), Q(1), Q(2), Q(3), Q(4), Q(5), Q(6), Q(7), Q(8), Q(9), Q(10), Q(11), Q(12), Q(13), Q(14), Q(15),
#undef Q
   },

#if LINELET_MAX_LINES >= 16
   [4] = {
#define Q(i) [i] = { .quant = { \
   [ 0] = PSY_LINE(i - 4, 3, 2, 1, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 1] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 2] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 3] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 4] = PSY_LINE(i - 2, 1, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 5] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 6] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 7] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 8] = PSY_LINE(i - 3, 2, 1, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [ 9] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [10] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [11] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [12] = PSY_LINE(i - 2, 1, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [13] = PSY_LINE(i + 0, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [14] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET), \
   [15] = PSY_LINE(i - 1, 0, 0, 0, PSY_LUMA_OFFSET, PSY_CHROMA_OFFSET) }}
      Q(0), Q(1), Q(2), Q(3), Q(4), Q(5), Q(6), Q(7), Q(8), Q(9), Q(10), Q(11), Q(12), Q(13), Q(14), Q(15),
#undef Q
   },
#endif

};

struct linelet_rate_allocator
{
   struct linelet_precinct_range_level *ranges[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS];
   unsigned precinct_samples;
   unsigned precincts;
   unsigned lines;

   struct linelet_component components[LINELET_MAX_COMPONENTS];
   unsigned levels[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES];
   unsigned num_components;

   struct linelet_rate_allocator_quant quant;
   struct linelet_rate_allocator_quant *quants;
   unsigned num_quants;

   int min_q;
   int min_opt_q;

   uint8_t unary_cost[256];
};

static void linelet_rate_allocator_init_cost_table(linelet_rate_allocator_t *ra)
{
   for (int x = 0; x < 16; x++)
   {
      for (int y = 0; y < 16; y++)
      {
         unsigned cost;
         if (x == y)
            cost = 1;
         else
         {
            int pred = x - y;
            cost = pred ? (abs(pred) + 1 + !!y) : 1;
         }
         ra->unary_cost[(y << 4) | x] = cost;
      }
   }
}

void linelet_rate_allocator_set_quant(linelet_rate_allocator_t *ra,
      const struct linelet_rate_allocator_quant *quant, unsigned num_quant)
{
   linelet_free(ra->quants);
   ra->quants = linelet_calloc(num_quant, sizeof(*ra->quants));
   memcpy(ra->quants, quant, num_quant * sizeof(*ra->quants));
   ra->num_quants = num_quant;

   // Fixup so we don't have negative quants or too large quants.
   // It's helpful to allow this
   // when stamping out quantization matrices with macros.
   for (unsigned n = 0; n < num_quant; n++)
   {
      int8_t *data = &ra->quants[n].quant[0][0][0];
      for (unsigned i = 0; i < sizeof(ra->quants[n].quant); i++)
      {
         if (data[i] < 0)
            data[i] = 0;
         else if (data[i] > 14)
            data[i] = 14;
      }
   }
}

void linelet_rate_allocator_minimum_quant(linelet_rate_allocator_t *ra, int q)
{
   ra->min_q = q;
}

linelet_rate_allocator_t *linelet_rate_allocator_new(unsigned pixels_per_precinct,
      unsigned samples_per_line, unsigned lines,
      const struct linelet_component *components, unsigned num_components,
      const struct linelet_rate_allocator_quant *quant, unsigned num_quant)
{
   // Sanity check.
   if (sizeof(struct linelet_precinct_range_level) != sizeof(uint64_t))
      abort();
   ////

   linelet_rate_allocator_t *ra = linelet_calloc(1, sizeof(*ra));
   if (!ra)
      return NULL;

   ra->min_q = ra->min_opt_q = LINELET_RATE_ALLOCATOR_NO_QUANT;
   ra->precincts = samples_per_line / pixels_per_precinct;
   ra->precinct_samples = pixels_per_precinct;
   ra->num_components = num_components;
   ra->lines = lines;

   for (unsigned c = 0; c < num_components; c++)
      for (unsigned l = 0; l < lines; l++)
         for (unsigned b = 0; b < LINELET_MAX_BANDS; b++)
            if (!(ra->ranges[c][l][b] = linelet_calloc(ra->precincts, sizeof(struct linelet_precinct_range_level))))
               goto error;

   memcpy(ra->components, components, num_components * sizeof(*components));

   for (unsigned c = 0; c < num_components; c++)
      linelet_common_set_component_levels(ra->levels[c], lines, components[c].decomposition_levels);

   if (quant)
      linelet_rate_allocator_set_quant(ra, quant, num_quant);
   else
      linelet_rate_allocator_set_quant(ra, default_quant_table[linelet_log2i(ra->lines)], DEFAULT_QUANT_TABLE_ELEMS);

   linelet_rate_allocator_init_cost_table(ra);
   return ra;

error:
   linelet_rate_allocator_free(ra);
   return NULL;
}

void linelet_rate_allocator_free(linelet_rate_allocator_t *ra)
{
   if (!ra)
      return;

   for (unsigned c = 0; c < LINELET_MAX_COMPONENTS; c++)
      for (unsigned l = 0; l < LINELET_MAX_LINES; l++)
         for (unsigned b = 0; b < LINELET_MAX_BANDS; b++)
            linelet_free(ra->ranges[c][l][b]);

   linelet_free(ra->quants);
   linelet_free(ra);
}

static inline unsigned study_range_quantized(struct linelet_precinct_range_level *level, unsigned quant, unsigned samples_log2)
{
   int16_t add = (1 << (quant - 1)) - 1;
   int lo = (level->lo + add) >> quant;
   int hi = (level->hi + add) >> quant;

   unsigned bits = 0;
   if (lo || hi)
   {
      int pos_lo = ~lo;
      unsigned maximum = LLMAX(hi, pos_lo);

      bits = maximum ? ((sizeof(unsigned) * CHAR_BIT - linelet_clz(maximum)) + 1) : 1;
   }

   level->bits = bits;
   return level->bytes_total = ((bits << samples_log2) + 7) >> 3;
}

// Predictive signed unary code.
// Very simple.
static unsigned compute_precincts_overhead(const struct linelet_precinct_range_level *level,
      unsigned precincts, const uint8_t *unary_cost)
{
   uint8_t cur = 0;
   unsigned bits = 0;

   for (unsigned p = 0; p < precincts; p++)
   {
      cur = (cur << 4) | level[p].bits;
      bits += unary_cost[cur];
   }

   return (bits + 7) >> 3;
}

static unsigned study_ranges(struct linelet_precinct_range_level *level,
      const int16_t *src, unsigned precincts, unsigned step)
{
   linelet_ranges_min_max(level, src, precincts, step);

   TIMER_COUNTER(study_ranges);
   TIMER_START_COUNT(study_ranges, precincts * step);

   unsigned component_size = 0;
   for (unsigned p = 0; p < precincts; p++)
   {
      unsigned bits = 0;
      int lo = level[p].lo;
      int hi = level[p].hi;

      if (lo || hi)
      {
         lo = ~lo;
         unsigned maximum = LLMAX(lo, hi);
         bits = maximum ? ((sizeof(unsigned) * CHAR_BIT - linelet_clz(maximum)) + 1) : 1;
      }

      level[p].bits = bits;
      level[p].bytes_total = (bits * step + 7) >> 3;
      component_size += level[p].bytes_total;
   }

   TIMER_STOP(study_ranges);
   return component_size;
}

static int study_precinct_savings(linelet_rate_allocator_t *ra,
      struct linelet_precinct_range_level *ranges, unsigned *precincts_overhead, unsigned quant, unsigned samples_log2)
{
   int savings = 0;
   for (unsigned p = 0; p < ra->precincts; p++)
   {
      savings += ranges[p].bytes_total;
      savings -= study_range_quantized(&ranges[p], quant, samples_log2);
   }

   if (precincts_overhead)
   {
      savings += *precincts_overhead;
      savings -= *precincts_overhead = compute_precincts_overhead(ranges, ra->precincts, ra->unary_cost);
   }

   return savings;
}

static size_t linelet_rate_allocator_study_component_line(linelet_rate_allocator_t *ra,
      const struct linelet_wavelet_buffer *buffer,
      struct linelet_precinct_range_level **ranges,
      unsigned c,
      unsigned line)
{
   size_t component_size = 0; 
   unsigned levels = ra->levels[c][line];
   unsigned decimation = ra->components[c].decimation;

   for (unsigned l = 0; l <= levels; l++)
   {
      const int16_t *src = l == levels ? buffer->lp[line][levels - 1] : buffer->hp[line][l];
      unsigned level_shift = l == levels ? l : l + 1;

      unsigned step = ra->precinct_samples >> (decimation + level_shift);
      component_size += study_ranges(ranges[l], src, ra->precincts, step);
   }

   return component_size;
}

static unsigned study_quantized_savings(linelet_rate_allocator_t *ra, unsigned total_size, unsigned min_q)
{
   const struct linelet_rate_allocator_quant *quant = &ra->quants[min_q];
   unsigned precinct_samples_log2 = linelet_log2i(ra->precinct_samples);

   for (int c = ra->num_components - 1; c >= 0; c--)
   {
      unsigned decimation = ra->components[c].decimation;
      for (int line = ra->lines - 1; line >= 0; line--)
      {
         unsigned levels = ra->levels[c][line];
         const uint8_t *quants = (const uint8_t*)quant->quant[line][c];
         uint8_t *out_quants = (uint8_t*)ra->quant.quant[line][c];
         struct linelet_precinct_range_level **ranges = ra->ranges[c][line];

         for (unsigned l = 0; l <= levels; l++)
         {
            unsigned samples_log2 = precinct_samples_log2 - (decimation + l + ((l == levels) ? 0 : 1));
            unsigned Q = quants[l];
            if (Q > out_quants[l])
            {
               total_size -= study_precinct_savings(ra, ranges[l], NULL, Q, samples_log2);
               out_quants[l] = Q;
            }
         }
      }
   }

   return total_size;
}

static unsigned study_quantized_savings_target(linelet_rate_allocator_t *ra,
      unsigned total_size, unsigned target_size, unsigned min_q,
      unsigned precincts_overhead[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS])
{
   const struct linelet_rate_allocator_quant *quant = &ra->quants[min_q];
   unsigned precinct_samples_log2 = linelet_log2i(ra->precinct_samples);

   for (int c = ra->num_components - 1; c >= 0; c--)
   {
      unsigned decimation = ra->components[c].decimation;
      for (int line = ra->lines - 1; line >= 0; line--)
      {
         unsigned levels = ra->levels[c][line];
         const uint8_t *quants = (const uint8_t*)quant->quant[line][c];
         uint8_t *out_quants = (uint8_t*)ra->quant.quant[line][c];
         struct linelet_precinct_range_level **ranges = ra->ranges[c][line];

         for (unsigned l = 0; l <= levels; l++)
         {
            unsigned samples_log2 = precinct_samples_log2 - (decimation + l + ((l == levels) ? 0 : 1));
            unsigned Q = quants[l];
            if (Q > out_quants[l])
            {
               total_size -= study_precinct_savings(ra, ranges[l],
                     precincts_overhead ? &precincts_overhead[c][line][l] : NULL, Q, samples_log2);
               out_quants[l] = Q;
               if (total_size <= target_size)
                  return total_size;
            }
         }
      }
   }

   return total_size;
}

void linelet_rate_allocator_study_lines(linelet_rate_allocator_t *ra,
      const struct linelet_wavelet_buffer *buffers, unsigned target_buffer_size)
{
   if (!target_buffer_size) // Bogus
      return;

   unsigned total_size = 0;
   for (unsigned c = 0; c < ra->num_components; c++)
      for (unsigned l = 0; l < ra->lines; l++)
         total_size += linelet_rate_allocator_study_component_line(ra, &buffers[c], ra->ranges[c][l], c, l);

   memset(&ra->quant, 0, sizeof(ra->quant));

   // Enforce at least some degree of quantization.
   int start_q = LLMAX(ra->min_q, ra->min_opt_q);
   if (start_q >= 0)
      total_size = study_quantized_savings(ra, total_size, start_q);

   TIMER_COUNTER(study_quantized);
   TIMER_START_COUNT(study_quantized, ra->lines);

   // Aim for target without overhead calculation.
   // It is redundant to also check precinct overhead until we're starting to hit the rate we want anyways.
   while (total_size > target_buffer_size)
   {
      // Already computed for start_q, skip one.
      start_q++;
      total_size = study_quantized_savings_target(ra, total_size, target_buffer_size, start_q, NULL);
   }

   // Add overhead from precincts.
   unsigned precincts_overhead[LINELET_MAX_COMPONENTS][LINELET_MAX_LINES][LINELET_MAX_BANDS];
   for (unsigned c = 0; c < ra->num_components; c++)
   {
      for (unsigned line = 0; line < ra->lines; line++)
      {
         unsigned levels = ra->levels[c][line];
         for (unsigned l = 0; l <= levels; l++)
            total_size += precincts_overhead[c][line][l] = compute_precincts_overhead(ra->ranges[c][line][l], ra->precincts, ra->unary_cost);
      }
   }

   // Back up one since first pass might have stopped mid-loop.
   start_q = LLMAX(start_q - 1, -1);

   // Quantize a bit more if overhead from precincts went over buffer limit.
   while (total_size > target_buffer_size)
   {
      start_q++;
      total_size = study_quantized_savings_target(ra, total_size, target_buffer_size, start_q, precincts_overhead);
   }

   ra->min_opt_q = LLMAX(start_q - 2, -1);
   TIMER_STOP(study_quantized);
}

unsigned linelet_rate_allocator_quant_factor(linelet_rate_allocator_t *ra,
      unsigned component, unsigned line, unsigned band)
{
   return ra->quant.quant[line][component][band];
}

unsigned linelet_rate_allocator_bits(linelet_rate_allocator_t *ra,
      unsigned component, unsigned line, unsigned precinct, unsigned band)
{
   return ra->ranges[component][line][band][precinct].bits;
}

const struct linelet_precinct_range_level *linelet_rate_allocator_precincts(linelet_rate_allocator_t *ra, unsigned component, unsigned line, unsigned band)
{
   return ra->ranges[component][line][band];
}

