%include "inc/x86inc.asm"
%include "inc/x86util.asm"

%macro INIT_ADDER 2
   pcmpeqw %1, %1
   psrlw %1, 15
   psllw %1, %2
%endmacro

%macro INIT_MASK 2
   pcmpeqd %1, %1
   psrld %1, (32 - %2)
%endmacro

%macro INIT_MINMAX 3
   pxor %1, %1
   pcmpeqw %2, %2
   psrlw %2, (16 - %3)
%endmacro

%macro BT_YUV8_FORWARD 0
; void bt_yuv8_forward_line(int16_t *dst, const uint8_t *src, unsigned samples)
cglobal bt_yuv8_forward_line, 3, 3, 4
   INIT_ADDER m3, 7
   pxor m2, m2

.loop:
   mova m0, [r1]
   mova m1, m0
   punpcklbw m0, m2
   punpckhbw m1, m2
   psubw m0, m3
   psubw m1, m3

   mova [r0], m0
   mova [r0 + mmsize], m1

   add r0, mmsize * 2
   add r1, mmsize
   sub r2d, mmsize
   jg .loop
   
   RET
%endmacro

%macro BT_YUV8_INVERSE 0
; void bt_yuv8_inverse_line(uint8_t *dst, const int16_t *src, unsigned samples)
cglobal bt_yuv8_inverse_line, 3, 3, 4
   INIT_ADDER m3, 7

.loop:
   mova m0, [r1]
   mova m1, [r1 + mmsize]
   paddw m0, m3
   paddw m1, m3
   packuswb m0, m1
   mova [r0], m0

   add r0, mmsize
   add r1, mmsize * 2
   sub r2d, mmsize
   jg .loop
   
   RET
%endmacro

%macro BT_YUV16_FORWARD 1
; void bt_yuv10_forward_line(int16_t *dst, const uint16_t *src, unsigned samples)
cglobal bt_yuv%1_forward_line, 3, 3, 3
   sub r1, r0
   INIT_ADDER m2, (%1 - 1)

.loop:
   mova m0, [r0 + r1]
   mova m1, [r0 + r1 + mmsize]
   psubw m0, m2
   psubw m1, m2
   mova [r0], m0
   mova [r0 + mmsize], m1

   add r0, mmsize * 2
   sub r2d, mmsize
   jg .loop

   RET
%endmacro

%macro BT_YUV16_INVERSE 1
; void bt_yuv10_inverse_line(uint16_t *dst, const int16_t *src, unsigned samples)
cglobal bt_yuv%1_inverse_line, 3, 3, 5
   sub r1, r0
   INIT_ADDER m2, (%1 - 1)
   INIT_MINMAX m3, m4, %1

.loop:
   mova m0, [r0 + r1]
   mova m1, [r0 + r1 + mmsize]
   paddw m0, m2
   paddw m1, m2
   CLIPW m0, m3, m4
   CLIPW m1, m3, m4
   mova [r0], m0
   mova [r0 + mmsize], m1

   add r0, mmsize * 2
   sub r2d, mmsize
   jg .loop

   RET
%endmacro

%macro RCT_FORWARD 1
   mova m0, [r3]
   mova m1, [r3 + mmsize]
   mova m2, m0
   mova m3, m1
   mova m4, m0
   mova m5, m1

%if %1 == 10
   pand m0, m6
   pand m1, m6
   psrld m2, 10
   psrld m3, 10
   psrld m4, 20
   psrld m5, 20
   pand m2, m6
   pand m3, m6
   pand m4, m6
   pand m5, m6
%else
   pand m4, m6
   pand m5, m6
   psrld m0, 16
   psrld m1, 16
   psrld m2, 8
   psrld m3, 8
   pand m0, m6
   pand m1, m6
   pand m2, m6
   pand m3, m6
%endif

   packssdw m0, m1 ; R
   packssdw m2, m3 ; G
   packssdw m4, m5 ; B
   mova m1, m0 ; R
   mova m3, m2 ; G
   mova m5, m4 ; B

   ; Compute luma
   psllw m2, 1
   psubw m0, m3 ; Cr
   paddw m2, m1
   psubw m4, m3 ; Cb
   paddw m2, m5
   psraw m2, 2
   psubw m2, m7 ; DC shift

   mova [r0], m2
   mova [r0 + r1], m4
   mova [r0 + r2], m0
   add r0, mmsize
   add r3, mmsize * 2
   sub r4d, mmsize / 2
%endmacro

%macro RCT_FORWARD_A2BGR10_YUV10 0
; void rct_forward_a2bgr10_yuv10_line(int16_t *y, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples)
cglobal rct_forward_a2bgr10_yuv10_line, 5, 5, 8
   sub r2, r0
   sub r1, r0
   INIT_ADDER m7, 9
   INIT_MASK m6, 10

.loop:
   RCT_FORWARD 10
   jg .loop

   RET
%endmacro

%macro RCT_INVERSE_YUV10_A2BGR10 0
; void rct_inverse_yuv10_a2bgr10_line(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples)
cglobal rct_inverse_yuv10_a2bgr10_line, 5, 5, 8
   sub r3, r1
   sub r2, r1

   INIT_MINMAX m5, m6, 10
   INIT_ADDER m7, 9

.loop:
   mova m0, [r1]
   mova m1, [r1 + r2]
   mova m2, [r1 + r3]
   mova m3, m1
   mova m4, m2

   paddw m1, m2
   psraw m1, 2
   psubw m0, m1
   paddw m3, m0
   paddw m4, m0

   paddw m0, m7 ; G
   paddw m3, m7 ; B
   paddw m4, m7 ; R

   ; R
   CLIPW m4, m5, m6
   mova m2, m4
   punpcklwd m2, m5
   punpckhwd m4, m5

   ; G
   CLIPW m0, m5, m6
   mova m1, m0
   punpcklwd m0, m5
   punpckhwd m1, m5
   pslld m0, 10
   pslld m1, 10
   por m2, m0
   por m4, m1

   ; B
   CLIPW m3, m5, m6
   mova m1, m3
   punpcklwd m1, m5
   punpckhwd m3, m5
   pslld m1, 20
   pslld m3, 20
   por m2, m1
   por m4, m3

   mova [r0], m2
   mova [r0 + mmsize], m4

   add r0, mmsize * 2
   add r1, mmsize
   sub r4d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro RCT_FORWARD_ARGB8_YUV8 0
; void rct_forward_a2bgr10_yuv10_line(int16_t *y, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples)
cglobal rct_forward_argb8_yuv8_line, 5, 5, 8
   sub r2, r0
   sub r1, r0
   INIT_ADDER m7, 7
   INIT_MASK m6, 8

.loop:
   RCT_FORWARD 8
   jg .loop

   RET
%endmacro

%macro RCT_INVERSE_YUV8_ARGB8 0
; void rct_inverse_yuv10_a2bgr10_line(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples)
cglobal rct_inverse_yuv8_argb8_line, 5, 5, 8
   sub r3, r1
   sub r2, r1

   INIT_ADDER m7, 7

.loop:
   mova m0, [r1]
   mova m1, [r1 + r2]
   mova m2, [r1 + r3]
   mova m3, m1
   mova m4, m2

   paddw m1, m2
   psraw m1, 2
   psubw m0, m1
   paddw m3, m0
   paddw m4, m0

   paddw m0, m7 ; G
   paddw m3, m7 ; B
   paddw m4, m7 ; R

   packuswb m4, m4
   packuswb m0, m0
   packuswb m3, m3

   punpcklbw m3, m0 ; BG x8
   punpcklbw m4, m0 ; RX x8, Alpha is G

   mova m0, m3
   punpcklwd m0, m4 ; BGRX x4
   punpckhwd m3, m4 ; BGRX x4

   mova [r0], m0
   mova [r0 + mmsize], m3

   add r0, mmsize * 2
   add r1, mmsize
   sub r4d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro RCT_FORWARD_GBRP8 0
; void rct_forward_gbrp8_line(int16_t *l, int16_t *cb, int16_t *cr, const uint8_t *g, const uint8_t *b, const uint8_t *r, unsigned samples)
cglobal rct_forward_gbrp8_line, 7, 7, 12
   sub r1, r0
   sub r2, r0
   sub r4, r3
   sub r5, r3

   INIT_ADDER m7, 7
   pxor m6, m6

.loop:
   mova m0, [r3]
   mova m1, [r3 + r4]
   mova m2, [r3 + r5]

   mova m3, m0
   mova m4, m1
   mova m5, m2

   punpcklbw m0, m6
   punpckhbw m3, m6
   punpcklbw m1, m6
   punpckhbw m4, m6
   punpcklbw m2, m6
   punpckhbw m5, m6

   ; B
   mova xmm8, m1
   mova xmm9, m4
   psubw xmm8, m0
   psubw xmm9, m3

   ; R
   mova xmm10, m2
   mova xmm11, m5
   psubw xmm10, m0
   psubw xmm11, m3

   ; Compute luma
   psllw m0, 1
   psllw m3, 1
   paddw m0, m1
   paddw m3, m4
   paddw m0, m2
   paddw m3, m5
   psraw m0, 2
   psraw m3, 2
   psubw m0, m7
   psubw m3, m7

   mova [r0], m0
   mova [r0 + mmsize], m3
   mova [r0 + r1], xmm8
   mova [r0 + r1 + mmsize], xmm9
   mova [r0 + r2], xmm10
   mova [r0 + r2 + mmsize], xmm11

   add r0, mmsize * 2
   add r3, mmsize
   sub r6d, mmsize
   jg .loop

   RET
%endmacro

%macro RCT_INVERSE_GBRP8 0
; void rct_inverse_gbrp8_line(uint8_t *g, uint8_t *b, uint8_t *r, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples)
cglobal rct_inverse_gbrp8_line, 7, 7, 10
   sub r1, r0
   sub r2, r0
   sub r4, r3
   sub r5, r3

   INIT_ADDER m7, 7

.loop:
   mova m0, [r3]
   mova m1, [r3 + r4]
   mova m2, [r3 + r5]
   mova m3, [r3 + mmsize]
   mova m4, [r3 + r4 + mmsize]
   mova m5, [r3 + r5 + mmsize]

   mova xmm8, m1
   paddw xmm8, m2
   mova xmm9, m4
   paddw xmm9, m5
   psraw xmm8, 2
   psraw xmm9, 2

   psubw m0, xmm8
   psubw m3, xmm9
   paddw m0, m7
   paddw m3, m7

   paddw m1, m0
   paddw m2, m0
   paddw m4, m3
   paddw m5, m3

   packuswb m0, m3
   packuswb m1, m4
   packuswb m2, m5

   mova [r0], m0
   mova [r0 + r1], m1
   mova [r0 + r2], m2

   add r0, mmsize
   add r3, mmsize * 2
   sub r6d, mmsize
   jg .loop

   RET
%endmacro

%macro RCT_FORWARD_GBRP16 1
; void rct_forward_gbrp10_line(int16_t *l, int16_t *cb, int16_t *cr, const uint16_t *g, const uint16_t *b, const uint16_t *r, unsigned samples)
cglobal rct_forward_gbrp%1_line, 7, 7, 6
   sub r1, r0
   sub r2, r0
   sub r3, r0
   sub r4, r0
   sub r5, r0

   INIT_ADDER m5, (%1 - 1)

.loop:
   mova m0, [r0 + r3]
   mova m1, [r0 + r4]
   mova m2, [r0 + r5]

   ; B
   mova m3, m1
   psubw m3, m0

   ; R
   mova m4, m2
   psubw m4, m0

   ; Compute luma
   paddw m1, m2
   psllw m0, 1
   paddw m0, m1
   psraw m0, 2
   psubw m0, m5

   mova [r0], m0
   mova [r0 + r1], m3
   mova [r0 + r2], m4

   add r0, mmsize
   sub r6d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro RCT_INVERSE_GBRP16 1
; void rct_inverse_gbrp8_line(uint16_t *g, uint16_t *b, uint16_t *r, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples)
cglobal rct_inverse_gbrp%1_line, 7, 7, 8
   sub r1, r0
   sub r2, r0
   sub r3, r0
   sub r4, r0
   sub r5, r0

   INIT_ADDER m7, (%1 - 1)
   INIT_MINMAX m5, m6, %1

.loop:
   mova m0, [r0 + r3]
   mova m1, [r0 + r4]
   mova m2, [r0 + r5]

   mova m4, m1
   paddw m4, m2
   psraw m4, 2

   psubw m0, m4
   paddw m0, m7

   paddw m1, m0
   paddw m2, m0

   CLIPW m0, m5, m6
   CLIPW m1, m5, m6
   CLIPW m2, m5, m6

   mova [r0], m0
   mova [r0 + r1], m1
   mova [r0 + r2], m2

   add r0, mmsize
   sub r6d, mmsize / 2
   jg .loop

   RET
%endmacro


; SSE4 would allow pmovzxbw, but it's slower than just mov + unpack.
INIT_XMM sse2
BT_YUV8_FORWARD
BT_YUV8_INVERSE
BT_YUV16_FORWARD 10
BT_YUV16_INVERSE 10
BT_YUV16_FORWARD 12
BT_YUV16_INVERSE 12
RCT_FORWARD_A2BGR10_YUV10
RCT_INVERSE_YUV10_A2BGR10
RCT_FORWARD_ARGB8_YUV8
RCT_INVERSE_YUV8_ARGB8
%if ARCH_X86_64 == 1 ; Using more than 6 GPR, don't bother with lots of hacks to get it working on 32-bit.
RCT_FORWARD_GBRP8
RCT_INVERSE_GBRP8
RCT_FORWARD_GBRP16 10
RCT_INVERSE_GBRP16 10
%endif

INIT_MMX mmx2
BT_YUV8_FORWARD
BT_YUV8_INVERSE
BT_YUV16_FORWARD 10
BT_YUV16_INVERSE 10
BT_YUV16_FORWARD 12
BT_YUV16_INVERSE 12
RCT_FORWARD_A2BGR10_YUV10
RCT_INVERSE_YUV10_A2BGR10
RCT_FORWARD_ARGB8_YUV8
RCT_INVERSE_YUV8_ARGB8
