%include "inc/x86inc.asm"
%include "inc/x86util.asm"

%macro PACK_EVEN 3
%if cpuflag(sse4)
   pand %1, %3
   pand %2, %3
   packusdw %1, %2
%else
   pslld %1, 16
   pslld %2, 16
   psrad %1, 16
   psrad %2, 16
   packssdw %1, %2
%endif
%endmacro

%macro PACK_ODD 2
   psrad %1, 16
   psrad %2, 16
   packssdw %1, %2
%endmacro

%macro SHIFT_LEFT_ONE 2
   PALIGNR %1, %2, (mmsize - 2), %2
%endmacro

%macro INIT_TWO_ADDER 1
   pcmpeqd %1, %1
   psrlw %1, 15
   psllw %1, 1
%endmacro

%macro INIT_ONE_ADDER 1
   pcmpeqd %1, %1
   psrlw %1, 15
%endmacro

%macro LEGALL_LINE_FORWARD 0
; void legall_line_forward(int16_t *lp, int16_t *hp, const int16_t *in, unsigned samples)
cglobal legall_line_forward, 4, 4, 8
   sub r1, r0

%if cpuflag(sse4)
   pcmpeqd m4, m4
   psrld m4, 16
%endif
   ; m5 = last_hp_out
   INIT_TWO_ADDER m6

.loop:
   mova m0, [r2]
   mova m1, [r2 + mmsize]
   mova m2, m0
   mova m3, m1
   PACK_EVEN m0, m1, m4
   PACK_ODD m2, m3

   movu m1, [r2 + 2]
   movu m3, [r2 + 2 + mmsize]
   PACK_ODD m1, m3

   paddw m1, m0
   psraw m1, 1
   psubw m2, m1

   mova m7, m2
   SHIFT_LEFT_ONE m7, m5

   paddw m7, m2
   paddw m7, m6
   psraw m7, 2
   paddw m0, m7

   mova [r0], m0
   mova [r0 + r1], m2
   mova m5, m2 ; save last_hp_out

   add r0, mmsize
   add r2, mmsize * 2
   sub r3d, mmsize
   jg .loop
   
   RET
%endmacro

%macro LEGALL_LINE_INVERSE 0
; void legall_line_inverse(int16_t *out, const int16_t *lp, const int16_t *hp, unsigned samples)
cglobal legall_line_inverse, 4, 4, 8
   sub r2, r1
   INIT_TWO_ADDER m6
   pinsrw m5, [r0 - 4], 7 ; last_even_out

.loop:
   movu m0, [r1] ; lp_t
   mova m1, [r1 + r2 - 2] ; hp_s
   movu m2, [r1 + r2] ; hp_t

   paddw m2, m1
   paddw m2, m6
   psraw m2, 2 ; hp_t

   psubw m0, m2 ; out_even

   mova m7, m0

   SHIFT_LEFT_ONE m7, m5 ; out_s
   mova m5, m0

   paddw m7, m0
   psraw m7, 1
   paddw m1, m7 ; out_odd

   ; BUTTERFLY :3 :3
   mova m2, m1
   punpcklwd m1, m0
   punpckhwd m2, m0

   movu [r0 - 2], m1
   movu [r0 + mmsize - 2], m2

   add r0, mmsize * 2
   add r1, mmsize
   sub r3d, mmsize
   jg .loop
   
   RET
%endmacro

%macro LEGALL_VERT_LINE_FORWARD_TOP 0
; void legall_vert_line_forward_top(int16_t *lp0, int16_t *hp0, int16_t *lp1, unsigned samples)
cglobal legall_vert_line_forward_top, 4, 4, 4
   sub r2, r0
   sub r1, r0
   INIT_TWO_ADDER m3

.loop:
   mova m0, [r0]
   mova m1, [r0 + r1]
   mova m2, [r0 + r2]

   paddw m2, m0
   psraw m2, 1
   psubw m1, m2
   mova [r0 + r1], m1

   psllw m1, 1
   paddw m1, m3
   psraw m1, 2
   paddw m0, m1
   mova [r0], m0

   add r0, mmsize
   sub r3d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro LEGALL_VERT_LINE_FORWARD 0
; void legall_vert_line_forward(int16_t *hp0, int16_t *lp1, int16_t *hp1, int16_t *hp2, unsigned samples)
cglobal legall_vert_line_forward, 5, 5, 5
   sub r3, r0
   sub r2, r0
   sub r1, r0
   INIT_TWO_ADDER m4

.loop:
   mova m0, [r0]
   mova m1, [r0 + r1]
   mova m2, [r0 + r2]
   mova m3, [r0 + r3]

   paddw m3, m1
   psraw m3, 1
   psubw m2, m3

   paddw m0, m4
   paddw m0, m2
   psraw m0, 2
   paddw m1, m0

   mova [r0 + r2], m2
   mova [r0 + r1], m1

   add r0, mmsize
   sub r4d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro LEGALL_VERT_LINE_FORWARD_BOTTOM 0
; void legall_vert_line_forward_bottom(int16_t *hp0, int16_t *lp1, int16_t *hp1, unsigned samples)
cglobal legall_vert_line_forward_bottom, 4, 4, 4
   sub r2, r0
   sub r1, r0
   INIT_TWO_ADDER m3

.loop:
   mova m0, [r0]
   mova m1, [r0 + r1]
   mova m2, [r0 + r2]

   psubw m2, m1

   paddw m0, m3
   paddw m0, m2
   psraw m0, 2
   paddw m1, m0

   mova [r0 + r2], m2
   mova [r0 + r1], m1

   add r0, mmsize
   sub r3d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro LEGALL_VERT_LINE_INVERSE_TOP 0
; void legall_vert_line_inverse_top(int16_t *lp0, int16_t *hp0, unsigned samples)
cglobal legall_vert_line_inverse_top, 3, 3, 3
   sub r1, r0
   INIT_ONE_ADDER m2

.loop:
   mova m0, [r0]
   mova m1, [r0 + r1]

   paddw m1, m2
   psraw m1, 1
   psubw m0, m1
   mova [r0], m0

   add r0, mmsize
   sub r2d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro LEGALL_VERT_LINE_INVERSE 0
; void legall_vert_line_inverse(int16_t *lp0, int16_t *hp0, int16_t *lp1, int16_t *hp1, unsigned samples)
cglobal legall_vert_line_inverse, 5, 5, 5
   sub r3, r0
   sub r2, r0
   sub r1, r0
   INIT_TWO_ADDER m4

.loop:
   mova m0, [r0]
   mova m1, [r0 + r1]
   mova m2, [r0 + r2]
   mova m3, [r0 + r3]

   paddw m3, m1
   paddw m3, m4
   psraw m3, 2
   psubw m2, m3

   paddw m0, m2
   psraw m0, 1
   paddw m1, m0

   mova [r0 + r2], m2
   mova [r0 + r1], m1

   add r0, mmsize
   sub r4d, mmsize / 2
   jg .loop

   RET
%endmacro

%macro LEGALL_VERT_LINE_INVERSE_BOTTOM 0
; void legall_vert_line_inverse_bottom(int16_t *lp0, int16_t *hp0, unsigned samples)
cglobal legall_vert_line_inverse_bottom, 3, 3, 3
   sub r1, r0

.loop:
   mova m1, [r0 + r1]
   paddw m1, [r0]
   mova [r0 + r1], m1

   add r0, mmsize
   sub r2d, mmsize / 2
   jg .loop

   RET
%endmacro

INIT_XMM sse2
LEGALL_LINE_FORWARD
LEGALL_LINE_INVERSE
LEGALL_VERT_LINE_FORWARD_TOP
LEGALL_VERT_LINE_FORWARD
LEGALL_VERT_LINE_FORWARD_BOTTOM
LEGALL_VERT_LINE_INVERSE_TOP
LEGALL_VERT_LINE_INVERSE
LEGALL_VERT_LINE_INVERSE_BOTTOM
INIT_XMM ssse3
LEGALL_LINE_FORWARD
LEGALL_LINE_INVERSE
INIT_XMM sse4
LEGALL_LINE_FORWARD
INIT_MMX mmx2
LEGALL_LINE_FORWARD
LEGALL_LINE_INVERSE
LEGALL_VERT_LINE_FORWARD_TOP
LEGALL_VERT_LINE_FORWARD
LEGALL_VERT_LINE_FORWARD_BOTTOM
LEGALL_VERT_LINE_INVERSE_TOP
LEGALL_VERT_LINE_INVERSE
LEGALL_VERT_LINE_INVERSE_BOTTOM

