%include "inc/x86inc.asm"
%include "inc/x86util.asm"

%macro PACK16 4
%if %2 <= %1
   pmovmskb %3d, %4
%if %2 != 1 && %2 != 9
   psllw %4, 1
%endif
   mov byte [r0 + (%1 - %2)], %3b
   %if mmsize > 8
      shr %3d, 8
      mov byte [r0 + (%1 - %2) + %1], %3b
   %endif
%endif
%endmacro

%macro PACK8 4
%if %2 <= %1
   pmovmskb %3d, %4
   psllw %4, 1
   mov byte [r0 + (%1 - %2)], %3b
%endif
%endmacro

%macro PACK4 4
%if %2 > 0
   pmovmskb %3d, %4
   psllw %4, 2
   %if %2 == 1
      and %3b, 0fh
   %endif
   mov byte [r0 + ((%1 - %2) >> 1)], %3b
%endif
%endmacro

%macro PACK_LINE16 1
; uint8_t *pack_line16(uint8_t *output, const int16_t *input, unsigned samples)
cglobal pack_line16_%1, 3, 5, 8
   mov r4d, r2d ; PACK16 macro fails for r3, r4 due to sil/dil not being available w/o REX on 32-bit

%if %1 > 8
   pcmpeqw m4, m4
   psrlw m4, 8
%endif

.loop:
   mova m0, [r1]
   mova m1, [r1 + mmsize]

%if %1 > 8
   mova m2, m0
   mova m3, m1
   psraw m2, (%1 - 8)
   psraw m3, (%1 - 8)
   packsswb m2, m3 ; v_hi
%endif

%if %1 > 8
   pand m0, m4
   pand m1, m4
   packuswb m0, m1
%elif %1 == 8
   packsswb m0, m1
%else
   packsswb m0, m1
   psllw m0, (8 - %1) ; v_lo
%endif

   ; v_lo and v_hi are now appropriately set up for packing.
   ; Completely unroll the packing.

   PACK16 %1, 16, r2, m2
   PACK16 %1,  8, r3, m0
   PACK16 %1, 15, r2, m2
   PACK16 %1,  7, r3, m0
   PACK16 %1, 14, r2, m2
   PACK16 %1,  6, r3, m0
   PACK16 %1, 13, r2, m2
   PACK16 %1,  5, r3, m0
   PACK16 %1, 12, r2, m2
   PACK16 %1,  4, r3, m0
   PACK16 %1, 11, r2, m2
   PACK16 %1,  3, r3, m0
   PACK16 %1, 10, r2, m2
   PACK16 %1,  2, r3, m0
   PACK16 %1,  9, r2, m2
   PACK16 %1,  1, r3, m0

   add r1, mmsize * 2
   add r0, (mmsize / 8) * %1
   sub r4d, mmsize
   jg .loop

   mov rax, r0
   RET
%endmacro

%macro PACK_LINE8 1
; uint8_t *pack_line8(uint8_t *output, const int16_t *input)
cglobal pack_line8_%1, 2, 4, 6

%if %1 > 8
   pcmpeqw m4, m4
   psrlw m4, 8
%endif

   mova m0, [r1]
%if mmsize == 8
   mova m1, [r1 + mmsize]
%endif

%if %1 > 8
   mova m2, m0
   psraw m2, (%1 - 8)
   %if mmsize == 8
      mova m3, m1
      psraw m3, (%1 - 8)
      packsswb m2, m3 ; v_hi
   %else
      packsswb m2, m2 ; v_hi
   %endif
%endif

%if mmsize == 8
   %if %1 > 8
      pand m0, m4
      pand m1, m4
      packuswb m0, m1
   %elif %1 == 8
      packsswb m0, m1
   %else
      packsswb m0, m1
      psllw m0, (8 - %1) ; v_lo
   %endif
%else
   %if %1 > 8
      pand m0, m4
      packuswb m0, m0
   %elif %1 == 8
      packsswb m0, m0
   %else
      packsswb m0, m0
      psllw m0, (8 - %1) ; v_lo
   %endif
%endif

   ; v_lo and v_hi are now appropriately set up for packing.
   ; Completely unroll the packing.

   PACK8 %1, 16, r2, m2
   PACK8 %1,  8, r3, m0
   PACK8 %1, 15, r2, m2
   PACK8 %1,  7, r3, m0
   PACK8 %1, 14, r2, m2
   PACK8 %1,  6, r3, m0
   PACK8 %1, 13, r2, m2
   PACK8 %1,  5, r3, m0
   PACK8 %1, 12, r2, m2
   PACK8 %1,  4, r3, m0
   PACK8 %1, 11, r2, m2
   PACK8 %1,  3, r3, m0
   PACK8 %1, 10, r2, m2
   PACK8 %1,  2, r3, m0
   PACK8 %1,  9, r2, m2
   PACK8 %1,  1, r3, m0

   lea rax, [r0 + %1]
   RET
%endmacro

%macro PACK_LINE4 1
; uint8_t *pack_line4(uint8_t *output, const int16_t *input)
cglobal pack_line4_%1, 2, 4, 6

%if %1 > 8
   pcmpeqw m4, m4
   psrlw m4, 8
%endif

   movq m2, [r1]
   mova m0, m2

%if %1 > 8
   psraw m2, (%1 - 8)
%elif %1 < 8
   psllw m2, (8 - %1)
%endif
   packsswb m2, m2

%if %1 > 1
   mova m3, m2
   psllw m3, 1
   punpckldq m2, m3 ; Facilitate unpacking two bitplanes at once.
%endif

%if %1 > 8
   pand m0, m4
   packuswb m0, m0
   psllw m0, (16 - %1) ; Keep doing double unpack as much as possible unless we're doing final bit.
%endif
%if %1 > 9
   mova m1, m0
   psllw m1, 1
   punpckldq m0, m1 ; Facilitate unpacking two bitplanes at once.
%endif

   ; v_lo and v_hi are now appropriately set up for packing.
   ; Use a different packing approach from PACK8 and PACK16

   PACK4 %1, (%1 -  0), r2, m2
   PACK4 %1, (%1 -  2), r2, m2
   PACK4 %1, (%1 -  4), r2, m2
   PACK4 %1, (%1 -  6), r2, m2
   PACK4 %1, (%1 -  8), r3, m0
   PACK4 %1, (%1 - 10), r3, m0
   PACK4 %1, (%1 - 12), r3, m0
   PACK4 %1, (%1 - 14), r3, m0

   lea rax, [r0 + ((%1 + 1) >> 1)]
   RET
%endmacro

%macro UNPACK 3
%if %1 >= %2
   %if %2 != %1 && %2 != 8
      psllw %3, 1
   %endif
   movzx r4, byte [r1 + (%1 - %2)]
   %if mmsize == 8
      por %3, [r3 + 8 * r4]
   %else
      movq m2, [r3 + 8 * r4]
      por %3, m2
   %endif
%endif
%endmacro

%macro UNPACK4 3
%if %1 >= %2
   %if %2 != %1 && %2 != 8
      psllw %3, 1
   %endif
   movzx r3, byte [r1 + ((%1 - %2) >> 1)]
   %if ((%1 - %2) & 1) == 1
      shr r3, 4
   %else
      and r3, 0fh
   %endif
   %if mmsize == 8
      por %3, [r2 + 8 * r3]
   %else
      movd m2, [r2 + 8 * r3]
      por %3, m2
   %endif
%endif
%endmacro

%macro UNPACK_LINE8 1
; uint8_t *unpack_line8(int16_t *output, const uint8_t *input, unsigned samples, const uint64_t *lut);
cglobal unpack_line8_%1, 4, 5, 3
.loop
   pxor m0, m0
   pxor m1, m1

   UNPACK %1, 16, m1
   UNPACK %1, 15, m1
   UNPACK %1, 14, m1
   UNPACK %1, 13, m1
   UNPACK %1, 12, m1
   UNPACK %1, 11, m1
   UNPACK %1, 10, m1
   UNPACK %1,  9, m1
   UNPACK %1,  8, m0
   UNPACK %1,  7, m0
   UNPACK %1,  6, m0
   UNPACK %1,  5, m0
   UNPACK %1,  4, m0
   UNPACK %1,  3, m0
   UNPACK %1,  2, m0
   UNPACK %1,  1, m0

%if mmsize == 8
   mova m2, m0
   punpcklbw m0, m1
   punpckhbw m2, m1
   psllw m0, (16 - %1)
   psllw m2, (16 - %1)
   psraw m0, (16 - %1)
   psraw m2, (16 - %1)
   mova [r0], m0
   mova [r0 + mmsize], m2
%else
   punpcklbw m0, m1
   psllw m0, (16 - %1)
   psraw m0, (16 - %1)
   mova [r0], m0
%endif

   add r0, 16
   add r1, %1
   sub r2d, 8
   jg .loop
   
   mov rax, r1
   RET
%endmacro

%macro UNPACK_LINE4 1
; uint8_t *unpack_line4(int16_t *output, const uint8_t *input, const uint64_t *lut);
cglobal unpack_line4_%1, 3, 4, 3
   pxor m0, m0
   pxor m1, m1

   UNPACK4 %1, 16, m1
   UNPACK4 %1, 15, m1
   UNPACK4 %1, 14, m1
   UNPACK4 %1, 13, m1
   UNPACK4 %1, 12, m1
   UNPACK4 %1, 11, m1
   UNPACK4 %1, 10, m1
   UNPACK4 %1,  9, m1
   UNPACK4 %1,  8, m0
   UNPACK4 %1,  7, m0
   UNPACK4 %1,  6, m0
   UNPACK4 %1,  5, m0
   UNPACK4 %1,  4, m0
   UNPACK4 %1,  3, m0
   UNPACK4 %1,  2, m0
   UNPACK4 %1,  1, m0

   punpcklbw m0, m1
   psllw m0, (16 - %1)
   psraw m0, (16 - %1)
   movq [r0], m0

   lea rax, [r1 + ((%1 + 1) >> 1)]
   RET
%endmacro


%macro PACK_LINE 1
PACK_LINE%1 1
PACK_LINE%1 2
PACK_LINE%1 3
PACK_LINE%1 4
PACK_LINE%1 5
PACK_LINE%1 6
PACK_LINE%1 7
PACK_LINE%1 8
PACK_LINE%1 9
PACK_LINE%1 10
PACK_LINE%1 11
PACK_LINE%1 12
PACK_LINE%1 13
PACK_LINE%1 14
PACK_LINE%1 15
PACK_LINE%1 16
%endmacro

%macro UNPACK_LINE 1
UNPACK_LINE%1 1
UNPACK_LINE%1 2
UNPACK_LINE%1 3
UNPACK_LINE%1 4
UNPACK_LINE%1 5
UNPACK_LINE%1 6
UNPACK_LINE%1 7
UNPACK_LINE%1 8
UNPACK_LINE%1 9
UNPACK_LINE%1 10
UNPACK_LINE%1 11
UNPACK_LINE%1 12
UNPACK_LINE%1 13
UNPACK_LINE%1 14
UNPACK_LINE%1 15
UNPACK_LINE%1 16
%endmacro

INIT_XMM sse2
PACK_LINE 16
PACK_LINE 8
PACK_LINE 4
UNPACK_LINE 8
UNPACK_LINE 4

INIT_MMX mmx2
PACK_LINE 16
PACK_LINE 8
PACK_LINE 4
UNPACK_LINE 8
UNPACK_LINE 4

