#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "rate_allocator.h"
#include "simd.h"
#include "log.h"

#define LINELET_SIMD_MMX2 (1 << 0)
#define LINELET_SIMD_SSE2 (1 << 1)
#define LINELET_SIMD_SSSE3 (1 << 2)
#define LINELET_SIMD_SSE4 (1 << 3)

// ASM routines
// RCT
void linelet_rct_forward_argb8_yuv8_line_sse2(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_forward_argb8_yuv8_line_mmx2(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_inverse_yuv8_argb8_line_sse2(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);
void linelet_rct_inverse_yuv8_argb8_line_mmx2(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

void linelet_rct_forward_a2bgr10_yuv10_line_sse2(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_forward_a2bgr10_yuv10_line_mmx2(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_inverse_yuv10_a2bgr10_line_sse2(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);
void linelet_rct_inverse_yuv10_a2bgr10_line_mmx2(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

void linelet_rct_forward_gbrp8_line_sse2(int16_t *l, int16_t *cb, int16_t *cr, const uint8_t *g, const uint8_t *b, const uint8_t *r, unsigned samples);
void linelet_rct_inverse_gbrp8_line_sse2(uint8_t *g, uint8_t *b, uint8_t *r, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);
void linelet_rct_forward_gbrp10_line_sse2(int16_t *l, int16_t *cb, int16_t *cr, const uint16_t *g, const uint16_t *b, const uint16_t *r, unsigned samples);
void linelet_rct_inverse_gbrp10_line_sse2(uint16_t *g, uint16_t *b, uint16_t *r, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

void linelet_rct_forward_argb8_yuv8_line_sse2(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_inverse_yuv8_argb8_line_sse2(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

void linelet_rct_forward_a2bgr10_yuv10_line_sse2(int16_t *l, int16_t *cb, int16_t *cr, const uint32_t *src, unsigned samples);
void linelet_rct_inverse_yuv10_a2bgr10_line_sse2(uint32_t *dst, const int16_t *l, const int16_t *cb, const int16_t *cr, unsigned samples);

// BT
void linelet_bt_yuv8_forward_line_sse2(int16_t *dst, const uint8_t *src, unsigned samples);
void linelet_bt_yuv8_forward_line_mmx2(int16_t *dst, const uint8_t *src, unsigned samples);
void linelet_bt_yuv8_inverse_line_sse2(uint8_t *dst, const int16_t *src, unsigned samples);
void linelet_bt_yuv8_inverse_line_mmx2(uint8_t *dst, const int16_t *src, unsigned samples);

void linelet_bt_yuv10_forward_line_sse2(int16_t *dst, const uint16_t *src, unsigned samples);
void linelet_bt_yuv10_forward_line_mmx2(int16_t *dst, const uint16_t *src, unsigned samples);
void linelet_bt_yuv10_inverse_line_sse2(uint16_t *dst, const int16_t *src, unsigned samples);
void linelet_bt_yuv10_inverse_line_mmx2(uint16_t *dst, const int16_t *src, unsigned samples);

void linelet_bt_yuv12_forward_line_sse2(int16_t *dst, const uint16_t *src, unsigned samples);
void linelet_bt_yuv12_forward_line_mmx2(int16_t *dst, const uint16_t *src, unsigned samples);
void linelet_bt_yuv12_inverse_line_sse2(uint16_t *dst, const int16_t *src, unsigned samples);
void linelet_bt_yuv12_inverse_line_mmx2(uint16_t *dst, const int16_t *src, unsigned samples);

// Haar
void linelet_haar_line_forward_sse2(int16_t *a, int16_t *b, unsigned samples);
void linelet_haar_line_inverse_sse2(int16_t *a, int16_t *b, unsigned samples);
void linelet_haar_line_forward_mmx2(int16_t *a, int16_t *b, unsigned samples);
void linelet_haar_line_inverse_mmx2(int16_t *a, int16_t *b, unsigned samples);

// Le Gall
void linelet_legall_line_forward_sse2(int16_t *lp, int16_t *hp, const int16_t *in, unsigned samples);
void linelet_legall_line_forward_ssse3(int16_t *lp, int16_t *hp, const int16_t *in, unsigned samples);
void linelet_legall_line_forward_sse4(int16_t *lp, int16_t *hp, const int16_t *in, unsigned samples);
void linelet_legall_line_forward_mmx2(int16_t *lp, int16_t *hp, const int16_t *in, unsigned samples);

void linelet_legall_line_inverse_sse2(int16_t *out, const int16_t *lp, const int16_t *hp, unsigned samples);
void linelet_legall_line_inverse_ssse3(int16_t *out, const int16_t *lp, const int16_t *hp, unsigned samples);
void linelet_legall_line_inverse_mmx2(int16_t *out, const int16_t *lp, const int16_t *hp, unsigned samples);

void linelet_legall_vert_line_forward_top_mmx2(
      int16_t *lp0,
      int16_t *hp0,
      int16_t *lp1,
      unsigned samples);

void linelet_legall_vert_line_forward_mmx2(
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      int16_t *lp2,
      unsigned samples);

void linelet_legall_vert_line_forward_bottom_mmx2(
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      unsigned samples);

void linelet_legall_vert_line_inverse_top_mmx2(
      int16_t *lp0,
      int16_t *hp0,
      unsigned samples);

void linelet_legall_vert_line_inverse_mmx2(
      int16_t *lp0,
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      unsigned samples);

void linelet_legall_vert_line_inverse_bottom_mmx2(
      int16_t *lp0,
      int16_t *hp0,
      unsigned samples);

void linelet_legall_vert_line_forward_top_sse2(
      int16_t *lp0,
      int16_t *hp0,
      int16_t *lp1,
      unsigned samples);

void linelet_legall_vert_line_forward_sse2(
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      int16_t *lp2,
      unsigned samples);

void linelet_legall_vert_line_forward_bottom_sse2(
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      unsigned samples);

void linelet_legall_vert_line_inverse_top_sse2(
      int16_t *lp0,
      int16_t *hp0,
      unsigned samples);

void linelet_legall_vert_line_inverse_sse2(
      int16_t *lp0,
      int16_t *hp0,
      int16_t *lp1,
      int16_t *hp1,
      unsigned samples);

void linelet_legall_vert_line_inverse_bottom_sse2(
      int16_t *lp0,
      int16_t *hp0,
      unsigned samples);

// Prepost
void linelet_prefilter_line2_sse2(int16_t *l0, int16_t *l1, unsigned samples);
void linelet_postfilter_line2_sse2(int16_t *l0, int16_t *l1, unsigned samples);
void linelet_prefilter_line2_mmx2(int16_t *l0, int16_t *l1, unsigned samples);
void linelet_postfilter_line2_mmx2(int16_t *l0, int16_t *l1, unsigned samples);

void linelet_prefilter_line4_sse2(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples);
void linelet_postfilter_line4_sse2(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples);
void linelet_prefilter_line4_mmx2(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples);
void linelet_postfilter_line4_mmx2(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, unsigned samples);

void linelet_prefilter_line8_sse2(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, int16_t *l4, int16_t *l5, int16_t *l6, int16_t *l7, unsigned samples);
void linelet_postfilter_line8_sse2(int16_t *l0, int16_t *l1, int16_t *l2, int16_t *l3, int16_t *l4, int16_t *l5, int16_t *l6, int16_t *l7, unsigned samples);


// Range
void linelet_range_min_max_line_8_sse2(struct linelet_precinct_range_level *ranges, const int16_t *src, unsigned precincts, unsigned step);
void linelet_range_min_max_line_8_sse4(struct linelet_precinct_range_level *ranges, const int16_t *src, unsigned precincts, unsigned step);
void linelet_range_min_max_line_8_mmx2(struct linelet_precinct_range_level *ranges, const int16_t *src, unsigned precincts, unsigned step);
void linelet_range_min_max_line_4_sse2(struct linelet_precinct_range_level *ranges, const int16_t *src, unsigned precincts);
void linelet_range_min_max_line_4_mmx2(struct linelet_precinct_range_level *ranges, const int16_t *src, unsigned precincts);

// Quantizer
void linelet_quantize_line_sse2(int16_t *data, unsigned samples, unsigned q, int16_t adder);
void linelet_dequantize_line_sse2(int16_t *data, unsigned samples, unsigned q);
void linelet_quantize_line_mmx2(int16_t *data, unsigned samples, unsigned q, int16_t adder);
void linelet_dequantize_line_mmx2(int16_t *data, unsigned samples, unsigned q);

// Packing
#define PACK16(v, inst) uint8_t *linelet_pack_line16_##v##_##inst(uint8_t *output, const int16_t *input, unsigned samples)
#define PACK8(v, inst) uint8_t *linelet_pack_line8_##v##_##inst(uint8_t *output, const int16_t *input)
#define PACK4(v, inst) uint8_t *linelet_pack_line4_##v##_##inst(uint8_t *output, const int16_t *input)
#define UNPACK8(v, inst) const uint8_t *linelet_unpack_line8_##v##_##inst(int16_t *output, const uint8_t *input, unsigned samples, const uint64_t *lut)
#define UNPACK4(v, inst) const uint8_t *linelet_unpack_line4_##v##_##inst(int16_t *output, const uint8_t *input, const uint64_t *lut)

#define STAMP_DECL(func, inst) \
   func(1, inst); func(2, inst); func(3, inst); func(4, inst); \
   func(5, inst); func(6, inst); func(7, inst); func(8, inst); \
   func(9, inst); func(10, inst); func(11, inst); func(12, inst); \
   func(13, inst); func(14, inst); func(15, inst); func(16, inst)

STAMP_DECL(PACK16,  mmx2);
STAMP_DECL(PACK16,  sse2);
STAMP_DECL(PACK8,   mmx2);
STAMP_DECL(PACK8,   sse2);
STAMP_DECL(PACK4,   mmx2);
STAMP_DECL(PACK4,   sse2);
STAMP_DECL(UNPACK8, mmx2);
STAMP_DECL(UNPACK8, sse2);
STAMP_DECL(UNPACK4, mmx2);
STAMP_DECL(UNPACK4, sse2);

#define FUNC(fun, len, v, inst) linelet_##fun##_line##len##_##v##_##inst

#define STAMP_TABLE(func, len, inst) \
   FUNC(func, len,  1, inst), \
   FUNC(func, len,  2, inst), \
   FUNC(func, len,  3, inst), \
   FUNC(func, len,  4, inst), \
   FUNC(func, len,  5, inst), \
   FUNC(func, len,  6, inst), \
   FUNC(func, len,  7, inst), \
   FUNC(func, len,  8, inst), \
   FUNC(func, len,  9, inst), \
   FUNC(func, len, 10, inst), \
   FUNC(func, len, 11, inst), \
   FUNC(func, len, 12, inst), \
   FUNC(func, len, 13, inst), \
   FUNC(func, len, 14, inst), \
   FUNC(func, len, 15, inst), \
   FUNC(func, len, 16, inst),

void linelet_x86_emms(void); // MMX2 crap.
void linelet_x86_cpuid(int flag, int flags[4]);
//

static const linelet_pack16_t pack16_sse2[16] = {
   STAMP_TABLE(pack, 16, sse2)
};

static const linelet_pack16_t pack16_mmx2[16] = {
   STAMP_TABLE(pack, 16, mmx2)
};

static const linelet_pack8_t pack8_sse2[16] = {
   STAMP_TABLE(pack, 8, sse2)
};

static const linelet_pack8_t pack8_mmx2[16] = {
   STAMP_TABLE(pack, 8, mmx2)
};

static const linelet_pack4_t pack4_sse2[16] = {
   STAMP_TABLE(pack, 4, sse2)
};

static const linelet_pack4_t pack4_mmx2[16] = {
   STAMP_TABLE(pack, 4, mmx2)
};

static const linelet_unpack8_t unpack8_sse2[16] = {
   STAMP_TABLE(unpack, 8, sse2)
};

static const linelet_unpack8_t unpack8_mmx2[16] = {
   STAMP_TABLE(unpack, 8, mmx2)
};

static const linelet_unpack4_t unpack4_sse2[16] = {
   STAMP_TABLE(unpack, 4, sse2)
};

static const linelet_unpack4_t unpack4_mmx2[16] = {
   STAMP_TABLE(unpack, 4, mmx2)
};

#undef INIT
#define INIT(func) linelet_simd.func = linelet_##func##_mmx2
static void init_mmx2(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "Enabling MMX2 optimizations.\n");
   linelet_simd.emms = linelet_x86_emms;

   INIT(prefilter_line2);
   INIT(postfilter_line2);
   INIT(prefilter_line4);
   INIT(postfilter_line4);

   INIT(rct_forward_argb8_yuv8_line);
   INIT(rct_forward_a2bgr10_yuv10_line);
   INIT(rct_inverse_yuv8_argb8_line);
   INIT(rct_inverse_yuv10_a2bgr10_line);

   INIT(bt_yuv8_forward_line);
   INIT(bt_yuv8_inverse_line);
   INIT(bt_yuv10_forward_line);
   INIT(bt_yuv10_inverse_line);
   INIT(bt_yuv12_forward_line);
   INIT(bt_yuv12_inverse_line);

   INIT(haar_line_forward);
   INIT(haar_line_inverse);

   INIT(legall_line_forward);
   INIT(legall_line_inverse);
   INIT(legall_vert_line_forward_top);
   INIT(legall_vert_line_forward);
   INIT(legall_vert_line_forward_bottom);
   INIT(legall_vert_line_inverse_top);
   INIT(legall_vert_line_inverse);
   INIT(legall_vert_line_inverse_bottom);

   INIT(range_min_max_line_8);
   INIT(range_min_max_line_4);

   INIT(quantize_line);
   INIT(dequantize_line);

   memcpy(linelet_simd.pack16,  pack16_mmx2,  sizeof(pack16_mmx2));
   memcpy(linelet_simd.pack8,   pack8_mmx2,   sizeof(pack8_mmx2));
   memcpy(linelet_simd.pack4,   pack4_mmx2,   sizeof(pack4_mmx2));
   memcpy(linelet_simd.unpack8, unpack8_mmx2, sizeof(unpack8_mmx2));
   memcpy(linelet_simd.unpack4, unpack4_mmx2, sizeof(unpack4_mmx2));
}

#undef INIT
#define INIT(func) linelet_simd.func = linelet_##func##_sse2
static void init_sse2(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "Enabling SSE2 optimizations.\n");

   INIT(prefilter_line2);
   INIT(postfilter_line2);
   INIT(prefilter_line4);
   INIT(postfilter_line4);
#ifdef __x86_64__
   INIT(prefilter_line8);
   INIT(postfilter_line8);
#endif

   INIT(rct_forward_argb8_yuv8_line);
   INIT(rct_forward_a2bgr10_yuv10_line);
   INIT(rct_inverse_yuv8_argb8_line);
   INIT(rct_inverse_yuv10_a2bgr10_line);

#ifdef __x86_64__
   INIT(rct_forward_gbrp8_line);
   INIT(rct_inverse_gbrp8_line);
   INIT(rct_forward_gbrp10_line);
   INIT(rct_inverse_gbrp10_line);
#endif

   INIT(bt_yuv8_forward_line);
   INIT(bt_yuv8_inverse_line);
   INIT(bt_yuv10_forward_line);
   INIT(bt_yuv10_inverse_line);
   INIT(bt_yuv12_forward_line);
   INIT(bt_yuv12_inverse_line);

   INIT(haar_line_forward);
   INIT(haar_line_inverse);

   INIT(legall_line_forward);
   INIT(legall_line_inverse);
   INIT(legall_vert_line_forward_top);
   INIT(legall_vert_line_forward);
   INIT(legall_vert_line_forward_bottom);
   INIT(legall_vert_line_inverse_top);
   INIT(legall_vert_line_inverse);
   INIT(legall_vert_line_inverse_bottom);

   INIT(range_min_max_line_8);
   INIT(range_min_max_line_4);

   INIT(quantize_line);
   INIT(dequantize_line);

   memcpy(linelet_simd.pack16,  pack16_sse2,  sizeof(pack16_sse2));
   memcpy(linelet_simd.pack8,   pack8_sse2,   sizeof(pack8_sse2));
   memcpy(linelet_simd.pack4,   pack4_sse2,   sizeof(pack4_sse2));
   memcpy(linelet_simd.unpack8, unpack8_sse2, sizeof(unpack8_sse2));
   memcpy(linelet_simd.unpack4, unpack4_sse2, sizeof(unpack4_sse2));
}

#undef INIT
#define INIT(func) linelet_simd.func = linelet_##func##_ssse3
static void init_ssse3(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "Enabling SSSE3 optimizations.\n");

   INIT(legall_line_forward);
   INIT(legall_line_inverse);
}

#undef INIT
#define INIT(func) linelet_simd.func = linelet_##func##_sse4
static void init_sse4(void)
{
   linelet_log_printf(LINELET_LOG_INFO, "Enabling SSE4 optimizations.\n");

   INIT(legall_line_forward);
   INIT(range_min_max_line_8);
}

void linelet_simd_init_x86(void)
{
   unsigned cpu = 0;
   int flags[4] = {0};
   linelet_x86_cpuid(0, flags);

   unsigned max_flag = flags[0];
   if (max_flag < 1) // Does CPUID not support func = 1? (unlikely ...)
   {
      linelet_log_printf(LINELET_LOG_INFO, "[CPUID]: Flag 0x1 not supported. (?!)\n");
      return;
   }

   linelet_x86_cpuid(1, flags);

   if (flags[3] & (1 << 25))
   {
      // SSE also implies MMXEXT (according to FFmpeg source).
      cpu |= LINELET_SIMD_MMX2;
   }

   if (flags[3] & (1 << 26))
      cpu |= LINELET_SIMD_SSE2;

   if (flags[2] & (1 << 9))
      cpu |= LINELET_SIMD_SSSE3;

   if (flags[2] & (1 << 19))
      cpu |= LINELET_SIMD_SSE4;

   linelet_x86_cpuid(0x80000000, flags);
   max_flag = flags[0];
   if (max_flag >= 0x80000001u)
   {
      linelet_x86_cpuid(0x80000001, flags);
      if (flags[3] & (1 << 22))
         cpu |= LINELET_SIMD_MMX2;
   }

   const char *env = getenv("LINELET_SIMD");
   if (env)
   {
      if (!strcmp(env, "mmx2"))
         init_mmx2();
      else if (!strcmp(env, "sse2"))
      {
         init_mmx2();
         init_sse2();
      }
      else if (!strcmp(env, "ssse3"))
      {
         init_mmx2();
         init_sse2();
         init_ssse3();
      }
      else if (!strcmp(env, "sse4"))
      {
         init_mmx2();
         init_sse2();
         init_ssse3();
         init_sse4();
      }
   }
   else
   {
      if (cpu & LINELET_SIMD_MMX2)
         init_mmx2();
      if (cpu & LINELET_SIMD_SSE2)
         init_sse2();
      if (cpu & LINELET_SIMD_SSSE3)
         init_ssse3();
      if (cpu & LINELET_SIMD_SSE4)
         init_sse4();
   }
}

