#!/usr/bin/env python

import sys
import os
import random
import subprocess
from time import gmtime, strftime

def test_paths(f, rate):
    base = os.path.splitext(f)[0].split('_lossless')[0]
    lossless_path = f
    lossy_path = base + '_' + str(rate)[:-3] + '.llv' if rate != 0 else lossless_path
    return (lossless_path, lossy_path)

def paths_to_command(lossless, lossy, gray_3, gray_8):
    return [lossless, gray_3, lossy, gray_3, lossless, gray_3, lossy, gray_8]

def list_to_int(buf):
    ret = [0]
    for rate in buf:
        ret.append(int(rate) * 1000)
    return ret

if __name__ == '__main__':
    bitrates = list_to_int(sys.argv[1].split(','))
    gray_3 = sys.argv[2]
    gray_8 = sys.argv[3]
    tests = sys.argv[4:]

    random.seed()
    rates_per_file = []
    for i in range(len(tests)):
        rates = bitrates[:]
        random.shuffle(rates)
        rates_per_file.append(rates)

    command = ['mpv']

    current = strftime('%Y-%m-%d-%H:%M:%S', gmtime())
    with open('linelet_test_log_' + current + '.txt', 'w') as f:
        for j in range(len(bitrates)):
            for i in range(len(tests)):
                rate = rates_per_file[i][j]
                test = tests[i]
                cmd = test_paths(test, rate)
                command.extend(paths_to_command(cmd[0], cmd[1], gray_3, gray_8))
                print(test, '@', (str(rate)[:-3] + 'Mbit/s') if rate != 0 else 'Lossless', file = f)

    # Run test
    subprocess.check_call(command)
