#include "tests.h"
#include "memory.h"
#include <stdint.h>
#include <string.h>

int convert_test(void)
{
   const uint32_t test_rgb[32] ALIGNED_STACK = {
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
      0xff542345, 0xffa6b6f6, 0xff8b1ac9, 0xff145244,
   };

   // RCT 8-bit
   uint32_t out_rgb[32] ALIGNED_STACK;
   int16_t l[32] ALIGNED_STACK, cb[32] ALIGNED_STACK, cr[32] ALIGNED_STACK;

   linelet_rct_forward_argb8888((int16_t*[]) { l, cb, cr }, 32, test_rgb, 16, 2, 64);
   linelet_rct_inverse_argb8888(out_rgb, 64, (const int16_t*[]) { l, cb, cr }, 16, 2, 32);
   for (unsigned i = 0; i < 32; i++)
      out_rgb[i] |= 0xff000000u;

   if (memcmp(out_rgb, test_rgb, sizeof(test_rgb)))
      return -1;

   // RCT 10-bit
   linelet_rct_forward_a2bgr10_yuv10((int16_t*[]) { l, cb, cr }, 32, test_rgb, 16, 2, 64);
   linelet_rct_inverse_yuv10_a2bgr10(out_rgb, 64, (const int16_t*[]) { l, cb, cr }, 16, 2, 32);
   for (unsigned i = 0; i < 32; i++)
      out_rgb[i] |= 0xc0000000u;
   if (memcmp(out_rgb, test_rgb, sizeof(test_rgb)))
      return -2;

   // BT709 8-bit
   const uint8_t test_luma_8[64] ALIGNED_STACK = {
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
   };
   const uint8_t test_cb_8[32] ALIGNED_STACK = {
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x93, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
   };
   const uint8_t test_cr_8[32] ALIGNED_STACK = {
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x20, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x40, 0x80, 0x33, 0x11,
   };

   int16_t l_8[64] ALIGNED_STACK, cb_8[32] ALIGNED_STACK, cr_8[32] ALIGNED_STACK;
   uint8_t out_l_8[64] ALIGNED_STACK, out_cb_8[32] ALIGNED_STACK, out_cr_8[32] ALIGNED_STACK;

   size_t strides_8[] = { 32, 16, 16 };
   size_t strides_16[] = { 64, 32, 32 };
   linelet_bt_yuv8_forward_dc_shift((int16_t*[]) { l_8, cb_8, cr_8 }, strides_16, (const uint8_t*[]) { test_luma_8, test_cb_8, test_cr_8 },
         strides_8, 2, strides_8);
   linelet_bt_yuv8_inverse_dc_shift((uint8_t*[]) { out_l_8, out_cb_8, out_cr_8 }, strides_8, (const int16_t*[]) { l_8, cb_8, cr_8 },
         strides_8, 2, strides_16);
   if (memcmp(test_luma_8, out_l_8, sizeof(out_l_8)))
      return -3;
   if (memcmp(test_cb_8, out_cb_8, sizeof(out_cb_8)))
      return -4;
   if (memcmp(test_cr_8, out_cr_8, sizeof(out_cr_8)))
      return -5;

   // GBRP 8-bit
   linelet_rct_forward_gbrp8((int16_t*[]) { l_8, cb_8, cr_8 }, strides_16[1], (const uint8_t*[]) { test_luma_8, test_cb_8, test_cr_8 },
         strides_8[1], 2, strides_8);
   linelet_rct_inverse_gbrp8((uint8_t*[]) { out_l_8, out_cb_8, out_cr_8 }, strides_8, (const int16_t*[]) { l_8, cb_8, cr_8 },
         strides_8[1], 2, strides_16[1]);
   if (memcmp(test_luma_8, out_l_8, sizeof(out_l_8)))
      return -6;
   if (memcmp(test_cb_8, out_cb_8, sizeof(out_cb_8)))
      return -7;
   if (memcmp(test_cr_8, out_cr_8, sizeof(out_cr_8)))
      return -8;

   // BT709 10/12-bit
   const uint16_t test_luma_10[64] ALIGNED_STACK = {
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x3ff, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x81, 0x3ff, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x81, 0x00, 0x10, 0x80, 0x33, 0x11,
   };
   const uint16_t test_cb_10[32] ALIGNED_STACK = {
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x93, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x3ff, 0x10, 0x80, 0x33, 0x11,
   };
   const uint16_t test_cr_10[32] ALIGNED_STACK = {
      0xff, 0x33, 0x11, 0x3ff, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x81, 0x3ff, 0x10, 0x80, 0x33, 0x11,
      0x2f, 0x33, 0x81, 0x3ff, 0x10, 0x80, 0x33, 0x11,
      0xff, 0x33, 0x11, 0x00, 0x40, 0x80, 0x33, 0x11,
   };

   int16_t l_10[64] ALIGNED_STACK, cb_10[32] ALIGNED_STACK, cr_10[32] ALIGNED_STACK;
   uint16_t out_l_10[64] ALIGNED_STACK, out_cb_10[32] ALIGNED_STACK, out_cr_10[32] ALIGNED_STACK;

   linelet_bt_yuv10_forward_dc_shift((int16_t*[]) { l_10, cb_10, cr_10 }, strides_16, (const uint16_t*[]) { test_luma_10, test_cb_10, test_cr_10 },
         strides_8, 2, strides_16);
   linelet_bt_yuv10_inverse_dc_shift((uint16_t*[]) { out_l_10, out_cb_10, out_cr_10 }, strides_16, (const int16_t*[]) { l_10, cb_10, cr_10 },
         strides_8, 2, strides_16);
   if (memcmp(test_luma_10, out_l_10, sizeof(out_l_10)))
      return -9;
   if (memcmp(test_cb_10, out_cb_10, sizeof(out_cb_10)))
      return -10;
   if (memcmp(test_cr_10, out_cr_10, sizeof(out_cr_10)))
      return -11;

   linelet_bt_yuv12_forward_dc_shift((int16_t*[]) { l_10, cb_10, cr_10 }, strides_16, (const uint16_t*[]) { test_luma_10, test_cb_10, test_cr_10 },
         strides_8, 2, strides_16);
   linelet_bt_yuv12_inverse_dc_shift((uint16_t*[]) { out_l_10, out_cb_10, out_cr_10 }, strides_16, (const int16_t*[]) { l_10, cb_10, cr_10 },
         strides_8, 2, strides_16);
   if (memcmp(test_luma_10, out_l_10, sizeof(out_l_10)))
      return -12;
   if (memcmp(test_cb_10, out_cb_10, sizeof(out_cb_10)))
      return -13;
   if (memcmp(test_cr_10, out_cr_10, sizeof(out_cr_10)))
      return -14;

   linelet_rct_forward_gbrp10((int16_t*[]) { l_10, cb_10, cr_10 }, strides_16[1], (const uint16_t*[]) { test_luma_10, test_cb_10, test_cr_10 },
         strides_8[1], 2, strides_16);
   linelet_rct_inverse_gbrp10((uint16_t*[]) { out_l_10, out_cb_10, out_cr_10 }, strides_16, (const int16_t*[]) { l_10, cb_10, cr_10 },
         strides_8[1], 2, strides_16[1]);
   if (memcmp(test_luma_10, out_l_10, sizeof(out_l_10)))
      return -15;
   if (memcmp(test_cb_10, out_cb_10, sizeof(out_cb_10)))
      return -16;
   if (memcmp(test_cr_10, out_cr_10, sizeof(out_cr_10)))
      return -17;

   return 0;
}

