#include "tests.h"
#include "memory.h"

int pack_u8(void)
{
   const uint8_t input[11] ALIGNED_STACK = {
      14, 9, 15, 3, 2, 5, 7, 6, 4, 1, 1,
   };

   uint8_t target[11] ALIGNED_STACK = {0};
   uint8_t output[11] ALIGNED_STACK = {0};

   unsigned bytes = (4 * 11 + 7) >> 3;
   unsigned ret = linelet_pack_nibbles_u8(target, input, 11);
   if (ret != bytes)
      return -1;

   ret = linelet_unpack_nibbles_u8(output, target, 11);
   if (ret != bytes)
      return -2;

   if (memcmp(output, input, sizeof(input)))
      return -3;

   return 0;
}

int pack_unary(void)
{
   const uint8_t input[] = {0, 2, 1, 3, 1, 1, 2, 2, 4, 5, 1, 2};
   uint8_t buffer[16];
   uint8_t output[sizeof(input)];

   size_t buffer_size = linelet_pack_predictive_unary(buffer, input, sizeof(input));
   size_t output_size = linelet_unpack_predictive_unary(output, buffer, sizeof(input));

   if (buffer_size != output_size)
      return -1;
   if (memcmp(input, output, sizeof(input)))
      return -2;
   return 0;
}

int pack_s16(void)
{
   const int16_t input[16][16] ALIGNED_STACK = {
      { 0, -1, 0, -1, 0, 0, -1, -1, 0, -1, 0, -1, 0, 0, -1, -1, },
      { 1, -2, 0, -2, 1, 1, 0, -2, 0, -1, 0, -1, 0, 0, -1, -1, },
      { 3, -4, -2, -4, 2, 0, 3, 2, 0, -1, 0, -1, 0, 0, -1, -1, },
      { 7, -8, -5, -2, -6, -7, 4, 5, 0, -1, 0, -1, 0, 0, -1, -1, },
      { 12, -14, -11, -16, 14, 11, 5, 2, 0, -1, 0, -1, 0, 0, -1, -1, },
      { 0, 2, 4, 6, 8, 10, 12, 14, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -64, 63, 62, -22, -45, 45, 22, 29, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -128, 127, 100, 80, 70, 50, -90, -110, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -256, 255, 200, 196, -225, -241, -56, 2, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -512, 511, -512, 511, 455, -11, 45, -167, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -1024, 1023, -1024, 1023, 955, -756, 45, -84, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -2048, 2047, -2048, 2047, 2012, -1874, -1431, -1111, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -4096, 4095, -4096, 4095, 2012, -1874, -1431, -1111, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -8092, 8000, -4096, 4095, 2012, -1874, -1431, -1111, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -16000, 16000, -4096, 4095, 2012, -1874, -1431, -1111, 0, -1, 0, -1, 0, 0, -1, -1, },
      { -32000, 32000, -4096, 4095, 2012, -1874, -1431, -1111, 0, -1, 0, -1, 0, 0, -1, -1, },
   };

   uint8_t target[16][32] ALIGNED_STACK = {{0}};
   int16_t output[16][16] ALIGNED_STACK = {{0}};

   int base_ret = 0;
   for (unsigned s = 2; s <= 16; s <<= 1)
   {
      for (unsigned b = 0; b < 16; b++, base_ret -= 10)
      {
         unsigned bytes = ((b + 1) * s + 7) >> 3;
         unsigned ret = linelet_pack_bits_s16(target[b], input[b], s, 1, (uint8_t[]) { b + 1 });
         if (ret != bytes)
            return base_ret - 1;

         ret = linelet_unpack_bits_s16(output[b], target[b], s, 1, (uint8_t[]) { b + 1 });
         if (ret != bytes)
            return base_ret - 2;

         if (memcmp(output[b], input[b], s * sizeof(int16_t)))
            return base_ret - 3;
      }
   }

   return 0;
}

