#include "tests.h"
#include "memory.h"

int rate_allocator_test(void)
{
   linelet_rate_allocator_t *ra = linelet_rate_allocator_new(8, 16, 1,
         (const struct linelet_component[]) {
            {
               .decomposition_levels = 1,
               .decimation = 0,
            },
            {
               .decomposition_levels = 1,
               .decimation = 1,
            },
            {
               .decomposition_levels = 1,
               .decimation = 1,
            },
         }, 3, NULL, 0);

   if (!ra)
      return -1;

   int16_t lp_buf_0[8] ALIGNED_STACK = { 32, -19, 55, -128, 12, 23, 34, 23 };
   int16_t hp_buf_0[8] ALIGNED_STACK = { 1, -1, 4, -1, -2, 5, 2, 0 };

   int16_t lp_buf_1[4] ALIGNED_STACK = { 32, -19, 55, -128 };
   int16_t hp_buf_1[4] ALIGNED_STACK = { 1, -1, 4, -1 };

   int16_t lp_buf_2[4] ALIGNED_STACK = { 32, -19, 55, -128 };
   int16_t hp_buf_2[4] ALIGNED_STACK = { 0, 0, 0, 0 };

   struct linelet_wavelet_buffer buffer[] = {
      {
         .lp = {{ lp_buf_0 }},
         .hp = {{ hp_buf_0 }},
      },
      {
         .lp = {{ lp_buf_1 }},
         .hp = {{ hp_buf_1 }},
      },
      {
         .lp = {{ lp_buf_2 }},
         .hp = {{ hp_buf_2 }},
      },
   };

   unsigned bits[3][2][2];
   unsigned quants[3][2];

   linelet_rate_allocator_study_lines(ra, buffer, 50);

   for (unsigned c = 0; c < 3; c++)
      for (unsigned p = 0; p < 2; p++)
         for (unsigned b = 0; b < 2; b++)
            bits[c][p][b] = linelet_rate_allocator_bits(ra, c, 0, p, b);

   for (unsigned c = 0; c < 3; c++)
      for (unsigned b = 0; b < 2; b++)
         quants[c][b] = linelet_rate_allocator_quant_factor(ra, c, 0, b);

   linelet_rate_allocator_free(ra);

   static const unsigned ref_bits_noquant[3][2][2] = {
      { { 4, 8 }, { 4, 7 } },
      { { 2, 7 }, { 4, 8 } },
      { { 0, 7 }, { 0, 8 } },
   };

   static const unsigned ref_quant_noquant[3][2] = {
      { 0, 0 },
      { 0, 0 },
      { 0, 0 },
   };

   if (memcmp(ref_bits_noquant, bits, sizeof(bits)))
      return -2;
   if (memcmp(ref_quant_noquant, quants, sizeof(quants)))
      return -3;

   return 0;
}

